/***************************************************************************
                           imdft.c  -  description
                             -------------------
   Implementation of the AC-3 IMDFT process.  Uses a decimation in time 
    split-radix FFT algorithm.
                             -------------------
    begin                : Mon Oct 24 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "imdft.h"

#define NUM_STAGES      7

#if defined(__MMX__) && (AC3_COEFF_SIZE == 16) && defined(_AC3_MMX_)

# include "mmx/mmx_imdft512.c"
# include "mmx/mmx_imdft256.c"

# if defined(__GCC__)
#  include "mmx/mmx_gcc_imdft512.c"
#  include "mmx/mmx_gcc_imdft256.c"
# elif defined(__MSVC__)
#  include "mmx/mmx_msvc_imdft512.c"
#  include "mmx/mmx_msvc_imdft256.c"
# else
#  error "Don't know how to insert inline ASM code for this compiler"
# endif

#else

#include "imdft_tables.h"

/*#define DISPLAY_LIMITS*/
#ifdef DISPLAY_LIMITS
# include <stdio.h>
static AC3_COEFF max_in  = AC3_COEFF_CONST(-1.0),
                 min_in  = AC3_COEFF_CONST( 1.0),
                 max_out = AC3_COEFF_CONST(-1.0),
                 min_out = AC3_COEFF_CONST( 1.0);
#endif

#define ERRORS 1

#define GET_OUTPUT(s, err) \
      tmp_i |= err; \
      tmp_r |= err; \
      /* x[1] = x[0] - tmp */ \
      *(img +IDX(s,1)) = *img  - tmp_i; \
      *(real-IDX(s,1)) = *real - tmp_r; \
      /* x[0] = x[0] + tmp */ \
      *img  += tmp_i; \
      *real += tmp_r; \
      /* tmp = x[3] */ \
      tmp_i = *(img +IDX(s,3)) | err; \
      tmp_r = *(real-IDX(s,3)) | err; \
      /* x[3] = x[2] - j * tmp */ \
      *(real-IDX(s,3)) = *(real-IDX(s,2)) + tmp_i; \
      *(img +IDX(s,3)) = *(img +IDX(s,2)) - tmp_r; \
      /* x[2] = x[2] + j * tmp */ \
      *(img +IDX(s,2)) += tmp_r; \
      *(real-IDX(s,2)) -= tmp_i

#define SR_BUTTERFLY_k_eq_0(s) \
      /* tmp = x[1] */ \
      tmp_i = *(img +IDX(s,3)); \
      tmp_r = *(real-IDX(s,3)); \
      /* x[3] = x[1] - x[3] */ \
      *(img +IDX(s,3)) = *(img +IDX(s,1)) - tmp_i; \
      *(real-IDX(s,3)) = *(real-IDX(s,1)) - tmp_r; \
      /* tmp = x[1] + x[3] */ \
      tmp_i += *(img +IDX(s,1)); \
      tmp_r += *(real-IDX(s,1)); \
      GET_OUTPUT(s, 0)

#define SR_BUTTERFLY_k_eq_N8(s) \
      img  += IDX(s,4); \
      real -= IDX(s,4); \
      /* tmp = (1+j) * x[1] */ \
      tmp_i = *(real-IDX(s,1)) + *(img +IDX(s,1)); \
      tmp_r = *(real-IDX(s,1)) - *(img +IDX(s,1)); \
      /* x[1] = (1-j) * x[3] */ \
      *(img +IDX(s,1)) = *(img +IDX(s,3)) - *(real-IDX(s,3)); \
      *(real-IDX(s,1)) = *(img +IDX(s,3)) + *(real-IDX(s,3)); \
      /* x[3] = ((1+j)*x[1] + (1-j)*x[3]) / sqrt(2) */ \
      *(img +IDX(s,3)) = tmp_i + *(img +IDX(s,1)); \
      *(real-IDX(s,3)) = tmp_r + *(real-IDX(s,1)); \
      *(img +IDX(s,3)) = AC3_COEFF_MULT(*(img +IDX(s,3)), \
                                        AC3_COEFF_CONST(0.707106781)); \
      *(real-IDX(s,3)) = AC3_COEFF_MULT(*(real-IDX(s,3)), \
                                        AC3_COEFF_CONST(0.707106781)); \
      /* tmp = ((1+j)*x[1] - (1-j)*x[3]) / sqrt(2) */ \
      tmp_i -= *(img +IDX(s,1)); \
      tmp_r -= *(real-IDX(s,1)); \
      tmp_i  = AC3_COEFF_MULT(tmp_i, AC3_COEFF_CONST(0.707106781)); \
      tmp_r  = AC3_COEFF_MULT(tmp_r, AC3_COEFF_CONST(0.707106781)); \
      GET_OUTPUT(s, ERRORS)

#define SR_BUTTERFLY(s) \
      img  += IDX(s,4); \
      real -= IDX(s,4); \
      /* tmp = W-kN * x[1] */ \
      tmp_i  = AC3_COEFF_MULT(*(img +IDX(s,1)), *idft); \
      tmp_r  = AC3_COEFF_MULT(*(real-IDX(s,1)), *(idft++)); \
      tmp_i += AC3_COEFF_MULT(*(real-IDX(s,1)), *idft); \
      tmp_r -= AC3_COEFF_MULT(*(img +IDX(s,1)), *(idft++)); \
      /* x[1] = W-3kN * x[3] */ \
      *(img +IDX(s,1))  = AC3_COEFF_MULT(*(img +IDX(s,3)), *idft); \
      *(real-IDX(s,1))  = AC3_COEFF_MULT(*(real-IDX(s,3)), *(idft++)); \
      *(img +IDX(s,1)) += AC3_COEFF_MULT(*(real-IDX(s,3)), *idft); \
      *(real-IDX(s,1)) -= AC3_COEFF_MULT(*(img +IDX(s,3)), *(idft++)); \
      /* x[3] = (W-kN * x[1]) - (W-3kN * x[3]) */ \
      *(img +IDX(s,3)) = tmp_i - *(img +IDX(s,1)); \
      *(real-IDX(s,3)) = tmp_r - *(real-IDX(s,1)); \
      /* tmp =  (W-kN * x[1]) + (W-3kN * x[3]) */ \
      tmp_i += *(img +IDX(s,1)) + (2*ERRORS); \
      tmp_r += *(real-IDX(s,1)); \
      GET_OUTPUT(s, 0)

static void pre_scale512(AC3_COEFF *real, AC3_COEFF *img);
static void pre_scale256(AC3_COEFF *real, AC3_COEFF *img);
static void post_scale512(AC3_COEFF *real, AC3_COEFF *img);
static void post_scale256(AC3_COEFF *real, AC3_COEFF *img);

void AC3_imdft512(AC3_COEFF *data)
{
   u32              cnt;
   AC3_COEFF        tmp_i, tmp_r;
   AC3_COEFF       *img  = data,
                   *real = data + ((2 << NUM_STAGES) - 1);
   const u8        *inc  = INC_PTR + INC_PTR_512;
   const AC3_COEFF *idft = IDFT_CONST;

   /* PRE IDFT STAGE */
   pre_scale512(real, img);

   /* STAGE 1: 2pt-DFTs */
   cnt = 1 << (NUM_STAGES-1);
   do {
/* ARRAY INDEX: 0 1 */
      /* tmp = x[1] */
      tmp_i = *(img +IDX(1,1));
      tmp_r = *(real-IDX(1,1));
      /* x[1] = x[0] - x[1] */
      *(img +IDX(1,1)) = *img  - tmp_i;
      *(real-IDX(1,1)) = *real - tmp_r;
      /* x[0] = x[0] + x[1] */
      *img  += tmp_i;
      *real += tmp_r;

      img  += 2;
      real -= 2;
/* OUTPUT ORDER: 0 1 */
   } while (--cnt);

   /* STAGE 2: 4pt-DFTs */
   img  -= 2 << (NUM_STAGES-1);
   real += 2 << (NUM_STAGES-1);
   goto START_STAGE_2;
   while (*inc) {
      img  += *inc;
      real -= *(inc++);
START_STAGE_2:
/* ARRAY INDEX: 0 1 2 3
    2-pt ORDER: 0   2
    2-pt ORDER:   1   3 */
      /* tmp = x[1] */
      tmp_i = *(img +IDX(2,1));
      tmp_r = *(real-IDX(2,1));
      GET_OUTPUT(2, 0);
/* k=0  (o0, o2, o1, o3) <= (i0, i1, i2, i3) */
/* OUTPUT ORDER: 0 2 1 3 */
   }

   /* STAGE 3: 8pt-DFTs */
   img  = data;
   real = data + ((2 << NUM_STAGES) - 1);
   goto START_STAGE_3;
   while (*(++inc)) {
      img  -= (u32)*inc;
      real += (u32)*inc;
START_STAGE_3:
/* ARRAY INDEX: 0 1 2 3 4 5 6 7
    4-pt ORDER: 0   4   2   6
    2-pt ORDER:   1       5
    2-pt ORDER:       3       7 */
      /* k=0 */
      SR_BUTTERFLY_k_eq_0(3);
      /* k=1 (N/8) */
      SR_BUTTERFLY_k_eq_N8(3);
/* k=0  (o0, o4, o2, o6) <= (i0, i1, i2, i3)
   k=1  (o1, o5, o3, o7) <= (i4, i5, i6, i7) */
/* OUTPUT ORDER: 0 4 2 6 1 5 3 7 */
   }

   /* STAGE 4: 16pt-DFTs */
   img  = data;
   real = data + ((2 << NUM_STAGES) - 1);
   goto START_STAGE_4;
   while (*(++inc)) {
      img  -= (u32)*inc;
      real += (u32)*inc;
START_STAGE_4:
/* ARRAY INDEX:  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15
    8-pt ORDER:  0     8     4    12     2    10     6    14
    4-pt ORDER:     1           9           5          13
    4-pt ORDER:           3          11          7           15 */
      /* k=0 */
      SR_BUTTERFLY_k_eq_0(4);
      /* k=2 (N/8) */
      SR_BUTTERFLY_k_eq_N8(4);
      /* k=1,3 */
      cnt = 2;
      do {
         SR_BUTTERFLY(4);
      } while (--cnt);
      idft -= 8;
/* k=0  ( o0,  o8,  o4, o12) <= ( i0,  i1,  i2,  i3)
   k=2  ( o2, o10,  o6, o14) <= ( i4,  i5,  i6,  i7)
   k=1  ( o1,  o9,  o5, o13) <= ( i8,  i9, i10, i11)
   k=3  ( o3, o11,  o7, o15) <= (i12, i13, i14, i15) */
/* OUTPUT ORDER: 0 8 4 12 2 10 6 14 1 9 5 13 3 11 7 15 */
   }
      
   /* STAGE 5: 32pt-DFTs */
   img  = data;
   real = data + ((2 << NUM_STAGES) - 1);
   goto START_STAGE_5;
   while (*(++inc)) {
      img  -= (u32)*inc;
      real += (u32)*inc;
START_STAGE_5:
      /* k=0 */
      SR_BUTTERFLY_k_eq_0(5);
      /* k=4 (N/8) */
      SR_BUTTERFLY_k_eq_N8(5);
      /* k=2,6,1,5,3,7 */
      cnt = 6;
      do {
         SR_BUTTERFLY(5);
      } while (--cnt);
      idft -= 24;
/* k=0  ( o0, o16,  o8, o24) <= ( i0,  i1,  i2,  i3)
   k=4  ( o4, o20, o12, o28) <= ( i4,  i5,  i6,  i7)
   k=2  ( o2, o18, o10, o26) <= ( i8,  i9, i10, i11)
   k=6  ( o6, o22, o14, o30) <= (i12, i13, i14, i15)
   k=1  ( o1, o17,  o9, o25) <= (i16, i17, i18, i19)
   k=5  ( o5, o21, o13, o29) <= (i20, i21, i22, i23)
   k=3  ( o3, o19, o11, o27) <= (i24, i25, i26, i27)
   k=7  ( o7, o23, o15, o31) <= (i28, i29, i30, i31) */
/* OUTPUT ORDER: 0 16 8 24 4 20 12 28 2 18 10 26 6 22 14 30
                 1 17 9 25 5 21 13 29 3 19 11 27 7 23 15 31 */
   }

   /* STAGE 6: 64pt-DFT */
   img  = data;
   real = data + ((2 << NUM_STAGES) - 1);
   {
      /* k=0 */
      SR_BUTTERFLY_k_eq_0(6);
      /* k=8 (N/8) */
      SR_BUTTERFLY_k_eq_N8(6);
      /* k=4,12,2,10,6,14,1,9,5,13,3,11,7,15 */
      cnt = 14;
      do {
         SR_BUTTERFLY(6);
      } while (--cnt);
      idft -= 56;
/* k= 0 ( o0, o32, o16, o48) <= ( i0,  i1,  i2,  i3)
   k= 8 ( o8, o40, o24, o56) <= ( i4,  i5,  i6,  i7)
   k= 4 ( o4, o36, o20, o52) <= ( i8,  i9, i10, i11)
   k=12 (o12, o44, o28, o60) <= (i12, i13, i14, i15)
   k= 2 ( o2, o34, o18, o50) <= (i16, i17, i18, i19)
   k=10 (o10, o42, o26, o58) <= (i20, i21, i22, i23)
   k= 6 ( o6, o38, o22, o54) <= (i24, i25, i26, i27)
   k=14 (o14, o46, o30, o62) <= (i28, i29, i30, i31)
   k= 1 ( o1, o33, o17, o49) <= (i32, i33, i34, i35)
   k= 9 ( o9, o41, o25, o57) <= (i36, i37, i38, i39)
   k= 5 ( o5, o37, o21, o53) <= (i40, i41, i42, i43)
   k=13 (o13, o45, o29, o61) <= (i44, i45, i46, i47)
   k= 3 ( o3, o35, o19, o51) <= (i48, i49, i50, i51)
   k=11 (o11, o43, o27, o59) <= (i52, i53, i54, i55)
   k= 7 ( o7, o39, o23, o55) <= (i56, i57, i58, i59)
   k=15 (o15, o47, o31, o63) <= (i60, i61, i62, i63) */
/* OUTPUT ORDER: 0 32 16 48  8 40 24 56 4 36 20 52 12 44 28 60
                 2 34 18 50 10 42 26 58 6 38 22 54 14 46 30 62
                 1 33 17 49  9 41 25 57 5 37 21 53 13 45 29 61
                 3 35 19 51 11 43 27 59 7 39 23 55 15 47 31 63 */
   }

   /* STAGE 7: 128pt-DFT */
   img  = data;
   real = data + ((2 << NUM_STAGES) - 1);
   {
      /* k=0 */
      SR_BUTTERFLY_k_eq_0(7);
      /* k=16 (N/8) */
      SR_BUTTERFLY_k_eq_N8(7);
      /* k=8,24,4,20,12,28,2,18,10,26,6,22,14,30,
           1,17,9,25,5,21,13,29,3,19,11,27,7,23,15,31 */
      cnt = 30;
      do {
         SR_BUTTERFLY(7);
      } while (--cnt);
/* k= 0 ( o0, o64, o32,  o96) <= (  i0,   i1,   i2,   i3)
   k=16 (o16, o80, o48, o112) <= (  i4,   i5,   i6,   i7)
   k= 8 ( o8, o72, o40, o104) <= (  i8,   i9,  i10,  i11)
   k=24 (o24, o88, o56, o120) <= ( i12,  i13,  i14,  i15)
   k= 4 ( o4, o68, o36, o100) <= ( i16,  i17,  i18,  i19)
   k=20 (o20, o84, o52, o116) <= ( i20,  i21,  i22,  i23)
   k=12 (o12, o76, o44, o108) <= ( i24,  i25,  i26,  i27)
   k=28 (o28, o92, o60, o124) <= ( i28,  i29,  i30,  i31)
   k= 2 ( o2, o66, o34,  o98) <= ( i32,  i33,  i34,  i35)
   k=18 (o18, o82, o50, o114) <= ( i36,  i37,  i38,  i39)
   k=10 (o10, o74, o42, o106) <= ( i40,  i41,  i42,  i43)
   k=26 (o26, o90, o58, o122) <= ( i44,  i45,  i46,  i47)
   k= 6 ( o6, o70, o38, o102) <= ( i48,  i49,  i50,  i51)
   k=22 (o22, o86, o54, o118) <= ( i52,  i53,  i54,  i55)
   k=14 (o14, o78, o46, o110) <= ( i56,  i57,  i58,  i59)
   k=30 (o30, o94, o62, o126) <= ( i60,  i61,  i62,  i63)
   k= 1 ( o1, o65, o33,  o97) <= ( i64,  i65,  i66,  i67)
   k=17 (o17, o81, o49, o113) <= ( i68,  i69,  i70,  i71)
   k= 9 ( o9, o73, o41, o105) <= ( i72,  i73,  i74,  i75)
   k=25 (o25, o89, o57, o121) <= ( i76,  i77,  i78,  i79)
   k= 5 ( o5, o69, o37, o101) <= ( i80,  i81,  i82,  i83)
   k=21 (o21, o85, o53, o117) <= ( i84,  i85,  i86,  i87)
   k=13 (o13, o77, o45, o109) <= ( i88,  i89,  i90,  i91)
   k=29 (o29, o93, o61, o125) <= ( i92,  i93,  i94,  i95)
   k= 3 ( o3, o67, o35,  o99) <= ( i96,  i97,  i98,  i99)
   k=19 (o19, o83, o51, o115) <= (i100, i101, i102, i103)
   k=11 (o11, o75, o43, o107) <= (i104, i105, i106, i107)
   k=27 (o27, o91, o59, o123) <= (i108, i109, i110, i111)
   k= 7 ( o7, o71, o39, o103) <= (i112, i113, i114, i115)
   k=23 (o23, o87, o55, o119) <= (i116, i117, i118, i119)
   k=15 (o15, o79, o47, o111) <= (i120, i121, i122, i123)
   k=31 (o31, o95, o63, o127) <= (i124, i125, i126, i127) */
/* OUTPUT ORDER: 0 64 32  96 16 80 48 112  8 72 40 104 24 88 56 120
                 4 68 36 100 20 84 52 116 12 76 44 108 28 92 60 124
                 2 66 34  98 18 82 50 114 10 74 42 106 26 90 58 122
                 6 70 38 102 22 86 54 118 14 78 46 110 30 94 62 126
                 
                 1 65 33  97 17 81 49 113  9 73 41 105 25 89 57 121
                 5 69 37 101 21 85 53 117 13 77 45 109 29 93 61 125
                 3 67 35  99 19 83 51 115 11 75 43 107 27 91 59 123
                 7 71 39 103 23 87 55 119 15 79 47 111 31 95 63 127 */
   }

   /* Undo bit-reverse order */
   img  = data;
   real = data + ((2 << NUM_STAGES) - 1);
   inc++;
   do {
      img  += *inc;
      real -= *(inc++);
      
      tmp_i          = *img;
      tmp_r          = *real;
      *img           = *(img  + *inc);
      *real          = *(real - *inc);
      *(img  + *inc) = tmp_i;
      *(real - *inc) = tmp_r;
   } while (*(++inc));

   /* POST IDFT STAGE */
   post_scale512(data + ((2 << NUM_STAGES)-1), data);
}

void AC3_imdft256_half(AC3_COEFF *data)
{
   u32              cnt;
   AC3_COEFF        tmp_i, tmp_r;
   AC3_COEFF       *img  = data, 
                   *real = data + ((2 << NUM_STAGES) - 2);
   const u8        *inc  = INC_PTR;
   const AC3_COEFF *idft = IDFT_CONST;

   pre_scale256(real, img);

   /* STAGE 1: 2pt-DFTs */
   cnt = 1 << (NUM_STAGES-2);
   do {
      tmp_i = *(img +IDX(1,1));
      tmp_r = *(real-IDX(1,1));
      *(img +IDX(1,1)) = *img  - tmp_i;
      *(real-IDX(1,1)) = *real - tmp_r;
      *img  += tmp_i;
      *real += tmp_r;
      img   += 4;
      real  -= 4;
/* OUTPUT ORDER: 0 1 */
   } while (--cnt);

   /* STAGE 2: 4pt-DFTs */
   img  -= 4 << (NUM_STAGES-2);
   real += 4 << (NUM_STAGES-2);
   goto START_STAGE_2b;
   while (*inc) {
      img  += *inc;
      real -= *(inc++);
START_STAGE_2b:
      tmp_i = *(img +IDX(2,1));
      tmp_r = *(real-IDX(2,1));
      GET_OUTPUT(2, 0);
   }

   /* STAGE 3: 8pt-DFTs */
   img  = data;
   real = data + ((2 << NUM_STAGES) - 2);
   goto START_STAGE_3b;
   while (*(++inc)) {
      img  -= (u32)*inc;
      real += (u32)*inc;
START_STAGE_3b:
      SR_BUTTERFLY_k_eq_0(3);
      SR_BUTTERFLY_k_eq_N8(3);
   }

   /* STAGE 4: 16pt-DFTs */
   img  = data;
   real = data + ((2 << NUM_STAGES) - 2);
   goto START_STAGE_4b;
   while (*(++inc)) {
      img  -= (u32)*inc;
      real += (u32)*inc;
START_STAGE_4b:
      SR_BUTTERFLY_k_eq_0(4);
      SR_BUTTERFLY_k_eq_N8(4);
      cnt = 2;
      do {
         SR_BUTTERFLY(4);
      } while (--cnt);
      idft -= 8;
   }

   /* STAGE 5: 32pt-DFTs */
   img  = data;
   real = data + ((2 << NUM_STAGES) - 2);
   goto START_STAGE_5b;
   while (*(++inc)) {
      img  -= (u32)*inc;
      real += (u32)*inc;
START_STAGE_5b:
      SR_BUTTERFLY_k_eq_0(5);
      SR_BUTTERFLY_k_eq_N8(5);
      cnt = 6;
      do {
         SR_BUTTERFLY(5);
      } while (--cnt);
      idft -= 24;
   }

   /* STAGE 6: 64pt-DFT */
   img  = data;
   real = data + ((2 << NUM_STAGES) - 2);
   {
      SR_BUTTERFLY_k_eq_0(6);
      SR_BUTTERFLY_k_eq_N8(6);
      cnt = 14;
      do {
         SR_BUTTERFLY(6);
      } while (--cnt);
      idft -= 56;
   }

   /* Undo bit-reverse order */
   img  = data;
   real = data + ((2 << NUM_STAGES) - 2);
   inc++;
   do {
      img  += *inc;
      real -= *(inc++);

      tmp_i          = *img;
      tmp_r          = *real;
      *img           = *(img  + *inc);
      *real          = *(real - *inc);
      *(img  + *inc) = tmp_i;
      *(real - *inc) = tmp_r;
   } while (*(++inc));
   
   post_scale256(data + ((2 << NUM_STAGES)-2), data);
}


static void pre_scale512(AC3_COEFF *real, AC3_COEFF *img)
{
   AC3_COEFF        tmp_i, tmp_r;
   const AC3_COEFF *scale = SCALE_FACTORS + (2 << NUM_STAGES);
   u32              cnt   = 1 << (NUM_STAGES-1);
# ifdef DISPLAY_LIMITS
   u32 change;
# endif
   
   do {
# ifdef DISPLAY_LIMITS
      change = 0;
      if (max_in < *img)
      {   max_in = *img; change = 1; }
      if (min_in > *img)
      {  min_in = *img; change = 1; }
      if (max_in < *real)
      {  max_in = *real; change = 1; }
      if (min_in > *real)
      {  min_in = *real; change = 1; }
      if (change)
         printf("max_in=0x%08X  min_in=0x%08X\n", (u32)max_in, (u32)min_in);
# endif
      tmp_i  = AC3_COEFF_MULT(*img , *scale);
      tmp_r  = AC3_COEFF_MULT(*real, *(scale++));
      tmp_i += AC3_COEFF_MULT(*real, *scale) + ERRORS;
      tmp_r -= AC3_COEFF_MULT(*img , *(scale++));
      if (cnt & 1) {
         *img  = -tmp_i;
         *real = -tmp_r;
      }
      else {
         *img  =  tmp_i;
         *real =  tmp_r;
      }
      img  += 2;
      real -= 2;
   } while (--cnt);

   scale -= 2 << NUM_STAGES;
   cnt    = 1 << (NUM_STAGES-1);
   do {
# ifdef DISPLAY_LIMITS
      change = 0;
      if (max_in < *img)
      {   max_in = *img; change = 1; }
      if (min_in > *img)
      {  min_in = *img; change = 1; }
      if (max_in < *real)
      {  max_in = *real; change = 1; }
      if (min_in > *real)
      {  min_in = *real; change = 1; }
      if (change)
         printf("max_in=0x%08X  min_in=0x%08X\n", (u32)max_in, (u32)min_in);
# endif
      tmp_r  = AC3_COEFF_MULT(*img , *scale);
      tmp_i  = AC3_COEFF_MULT(*real, *(scale++));
      tmp_r += AC3_COEFF_MULT(*real, *scale) + ERRORS;
      tmp_i -= AC3_COEFF_MULT(*img , *(scale++));
      if (cnt & 1) {
         *real =  tmp_r;
         *img  = -tmp_i;
      }
      else {
         *real = -tmp_r;
         *img  =  tmp_i;
      }
      real -= 2;
      img  += 2;
   } while (--cnt);
}

static void pre_scale256(AC3_COEFF *real, AC3_COEFF *img)
{
   AC3_COEFF        tmp_i, tmp_r;
   const AC3_COEFF *scale = SCALE_FACTORS;
   u32              cnt   = 1 << (NUM_STAGES-1);
# ifdef DISPLAY_LIMITS
   u32 change;
# endif
   
   do {
# ifdef DISPLAY_LIMITS
      change = 0;
      if (max_in < *img)
      {   max_in = *img; change = 1; }
      if (min_in > *img)
      {  min_in = *img; change = 1; }
      if (max_in < *real)
      {  max_in = *real; change = 1; }
      if (min_in > *real)
      {  min_in = *real; change = 1; }
      if (change)
         printf("max_in=0x%08X  min_in=0x%08X\n", (u32)max_in, (u32)min_in);
# endif
      tmp_i  = AC3_COEFF_MULT(*img , *scale);
      tmp_r  = AC3_COEFF_MULT(*real, *(scale++));
      tmp_i += AC3_COEFF_MULT(*real, *scale);
      tmp_r -= AC3_COEFF_MULT(*img , *(scale++));
      *img   = tmp_i + ERRORS;
      *real  = tmp_r;
      img  += 4;
      real -= 4;
   } while (--cnt);
}

static void post_scale512(AC3_COEFF *real, AC3_COEFF *img)
{
   AC3_COEFF        tmp_i, tmp_r;
   const AC3_COEFF *scale = SCALE_FACTORS + (2 << (NUM_STAGES-1));
   u32              cnt   = 1 << NUM_STAGES;
# ifdef DISPLAY_LIMITS
   u32 change;
# endif
   
   do {
      tmp_i  = AC3_COEFF_MULT(*img , *scale);
      tmp_r  = AC3_COEFF_MULT(*real, *(scale++));
      tmp_i += AC3_COEFF_MULT(*real, *scale);
      tmp_r -= AC3_COEFF_MULT(*img , *(scale++));
      *img   =  tmp_i + ERRORS;
      *real  = -tmp_r;
# ifdef DISPLAY_LIMITS
      change = 0;
      if (max_out < *img)
      {   max_out = *img; change = 1; }
      if (min_out > *img)
      {  min_out = *img; change = 1; }
      if (max_out < *real)
      {  max_out = *real; change = 1; }
      if (min_out > *real)
      {  min_out = *real; change = 1; }
      if (change)
         printf("max_out=0x%08X  min_out=0x%08X\n", (u32)max_out, (u32)min_out);
# endif
      img  += 2;
      real -= 2;
   } while (--cnt);
}

static void post_scale256(AC3_COEFF *real, AC3_COEFF *img)
{
   AC3_COEFF        tmp_i, tmp_r;
   const AC3_COEFF *scale = SCALE_FACTORS;
   u32              cnt   = 1 << (NUM_STAGES-1);
# ifdef DISPLAY_LIMITS
   u32 change;
# endif
   
   do {
      tmp_i  = AC3_COEFF_MULT(*img , *scale);
      tmp_r  = AC3_COEFF_MULT(*real, *(scale++));
      tmp_i += AC3_COEFF_MULT(*real, *scale);
      tmp_r -= AC3_COEFF_MULT(*img , *(scale++));
      *img   =  tmp_i + ERRORS;
      *real  = -tmp_r;
# ifdef DISPLAY_LIMITS
      change = 0;
      if (max_out < *img)
      {   max_out = *img; change = 1; }
      if (min_out > *img)
      {  min_out = *img; change = 1; }
      if (max_out < *real)
      {  max_out = *real; change = 1; }
      if (min_out > *real)
      {  min_out = *real; change = 1; }
      if (change)
         printf("max_out=0x%08X  min_out=0x%08X\n", (u32)max_out, (u32)min_out);
# endif
      img  += 4;
      real -= 4;
   } while (--cnt);
}

#undef ERRORS
#undef GET_OUTPUT
#undef SR_BUTTERFLY_k_eq_0
#undef SR_BUTTERFLY_k_eq_N8
#undef SR_BUTTERFLY

#endif

