/***************************************************************************
                           File.h  -  description
                             -------------------
   Wrapper for all file access operations.
                             -------------------
    begin                : Mon Aug 1 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __FILE_H__
#define __FILE_H__

#if defined(__WIN64_FILE__)||defined(__WIN32_FILE__)
# ifndef WIN32_LEAN_AND_MEAN
#  define WIN32_LEAN_AND_MEAN
# endif
# include <windows.h>
# include "typedefs.h"

typedef HANDLE FILE_PTR;
typedef DWORD  FILE_BYTE_COUNT;
# ifdef __WIN64_FILE__
typedef s64    FILE_POS;
# else
typedef DWORD  FILE_POS;
# endif

#define FILE_PTR_OK(ptr)      (ptr != INVALID_HANDLE_VALUE)
#define FILE_PTR_BAD(ptr)     (ptr == INVALID_HANDLE_VALUE)
#define FILE_PTR_BAD_VALUE    INVALID_HANDLE_VALUE

#define FILE_OPEN_READ(ptr, filename) \
   ptr = CreateFile(filename,\
                    GENERIC_READ,\
                    FILE_SHARE_READ,\
                    NULL,\
                    OPEN_EXISTING,\
                    FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN,\
                    NULL)
#define FILE_OPEN_WRITE(ptr, filename) \
   ptr = CreateFile(filename,\
                    GENERIC_WRITE,\
                    0,\
                    NULL,\
                    CREATE_ALWAYS,\
                    FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN,\
                    NULL)
#define FILE_OPEN_RDWR(ptr, filename) \
   ptr = CreateFile(filename,\
                    GENERIC_READ|GENERIC_WRITE,\
                    FILE_SHARE_READ,\
                    NULL,\
                    OPEN_ALWAYS,\
                    FILE_ATTRIBUTE_NORMAL|FILE_FLAG_RANDOM_ACCESS,\
                    NULL)

#define FILE_CLOSE(ptr)    CloseHandle(ptr)

#define FILE_READ(ptr, buf, buf_bytes, ret_bytes) \
   ReadFile(ptr, buf, buf_bytes, &(ret_bytes), NULL)
#define FILE_WRITE(ptr, buf, buf_bytes, ret_bytes) \
   WriteFile(ptr, buf, buf_bytes, &(ret_bytes), NULL)

#define FILE_FLUSH(ptr)    FlushFileBuffers(ptr)
#define FILE_REWIND(ptr)   SetFilePointer(ptr, 0, NULL, FILE_BEGIN)

# ifdef __WIN64_FILE__
#define FILE_SETPOS(ptr, pos) \
{ \
   LONG high_val = (LONG)((pos) >> 32); \
   SetFilePointer(ptr, (LONG)(pos), &high_val, FILE_BEGIN); \
}
#define FILE_GETPOS(ptr, pos) \
{ \
   LONG high_val = 0; \
   pos  = (s64)SetFilePointer(ptr, 0, &high_val, FILE_CURRENT); \
   pos |= (s64)high_val << 32; \
}
# else
#define FILE_SETPOS(ptr, pos) \
   SetFilePointer(ptr, (LONG)(pos), NULL, FILE_BEGIN)
#define FILE_GETPOS(ptr, pos) \
   pos = SetFilePointer(ptr, 0, NULL, FILE_CURRENT)
# endif

#elif defined(__UNISTD64_FILE__)
# include <fcntl.h>
# include <unistd.h>
# include <sys/types.h>

typedef int    FILE_PTR;
typedef off_t  FILE_POS;
typedef size_t FILE_BYTE_COUNT;

#define FILE_PTR_OK(ptr)      (ptr >= 0)
#define FILE_PTR_BAD(ptr)     (ptr < 0)
#define FILE_PTR_BAD_VALUE    -1

#define FILE_OPEN_READ(ptr, filename) \
   ptr = open(filename, O_RDONLY)
#define FILE_OPEN_WRITE(ptr, filename) \
   ptr = open(filename, (O_WRONLY|O_CREAT|O_TRUNC), 0644)
#define FILE_OPEN_RDWR(ptr, filename) \
   ptr = open(filename, (O_RDWR|O_CREAT), 0644)

#define FILE_CLOSE(ptr)    close(ptr)

#define FILE_READ(ptr, buf, buf_bytes, ret_bytes) \
   ret_bytes = read(ptr, buf, buf_bytes)
#define FILE_WRITE(ptr, buf, buf_bytes, ret_bytes) \
   ret_bytes = write(ptr, buf, buf_bytes)

#define FILE_FLUSH(ptr)
#define FILE_REWIND(ptr)         lseek(ptr, (off_t)0, SEEK_SET)
#define FILE_SETPOS(ptr, pos)    lseek(ptr, pos, SEEK_SET)
#define FILE_GETPOS(ptr, pos)    pos = lseek(ptr, (off_t)0, SEEK_CUR)

#else /* defined(__STDIO32_FILE__)||defined(__STDIO64_FILE__) */
# include <stdio.h>
# ifdef __STDIO64_FILE__
#  include <sys/types.h>

typedef off_t  FILE_POS;
# else
typedef long   FILE_POS;
# endif
typedef FILE * FILE_PTR;
typedef size_t FILE_BYTE_COUNT;

#define FILE_PTR_OK(ptr)      (ptr)
#define FILE_PTR_BAD(ptr)     (!ptr)
#define FILE_PTR_BAD_VALUE    NULL

#define FILE_OPEN_READ(ptr, filename)     ptr = fopen(filename, "rb")
#define FILE_OPEN_WRITE(ptr, filename)    ptr = fopen(filename, "wb")
#define FILE_OPEN_RDWR(ptr, filename)     ptr = fopen(filename, "rb+")

#define FILE_CLOSE(ptr)    fclose(ptr)

#define FILE_READ(ptr, buf, buf_bytes, ret_bytes) \
   ret_bytes = fread(buf, 1, buf_bytes, ptr)
#define FILE_WRITE(ptr, buf, buf_bytes, ret_bytes) \
   ret_bytes = fwrite(buf, 1, buf_bytes, ptr)

#define FILE_FLUSH(ptr)    fflush(ptr)
#define FILE_REWIND(ptr)   rewind(ptr)

# ifdef __STDIO64_FILE__
#  define FILE_GETPOS(ptr, pos)  pos = ftello(ptr)
#  define FILE_SETPOS(ptr, pos)  fseeko(ptr, pos, SEEK_SET)
# else
#  define FILE_GETPOS(ptr, pos)  pos = ftell(ptr)
#  define FILE_SETPOS(ptr, pos)  fseek(ptr, pos, SEEK_SET)
# endif

#endif

#endif

