/***************************************************************************
                          PMT_module.h  -  description
                             -------------------
   PMT PSI module functions.
                             -------------------
    begin                : Tue Aug 24 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __PMT_MODULE_H__
#define __PMT_MODULE_H__

#include "PAT_module.h"

/* Use this macro to initialize a MPEG2_PSI_PES_MODULE structure. */
#define PMT_PSI_MODULE \
{\
   PAT_PSI_MODULE_stream_init, \
   PMT_PSI_MODULE_process_pkt, \
   {\
      PMT_PSI_MODULE_mem_needed, \
      PMT_PSI_MODULE_init, \
      PMT_PSI_MODULE_done, \
      (MPEG2_MODULE_set_offset)NULL, \
      (MPEG2_MODULE_discontinuity)NULL, \
      NULL \
   }\
}

#ifdef __cplusplus
extern "C" {
#endif

/*****************************************************************
   GENERAL MODULE FUNCTIONS
     These functions are used by all SYSTEM, PSI, and PES modules.
 *****************************************************************/
/* Returns the number of bytes of memory needed by the module. */
size_t PMT_PSI_MODULE_mem_needed();

/* Initializes the memory provided for the module.
      INPUT: void *     - Pointer to the memory provided for the module.
             SYS_DATA * - Pointer to system data. */
void   PMT_PSI_MODULE_init(void *mem, SYS_DATA *sys_data);

/* Clean up (DO NOT FREE) the memory provided for the module.
      INPUT: void * - Pointer to the memory provided for the module. */
void   PMT_PSI_MODULE_done(void *mem);


/**************************************************************************
   COMMON PSI MODULE FUNCTIONS
     These functions are used by SYSTEM modules to process the PSI_SECTIONs
       they produce.
 **************************************************************************/
/* Processes the packet handled by the module.
      INPUT: void * - Pointer to the memory provided for the module.
             void * - Pointer to the PSI_SECTION that needs handling. */
void PMT_PSI_MODULE_process_pkt(void *mem, void *psi);


/*************************************************************************
   PMT MODULE FUNCTIONS
     These functions are used by the MPEG2 TS decoder to get the PMT data.
       The PAT data is extracted from the PSI_SECTIONs sent by a system
       module.
 *************************************************************************/
/* Returns a non-zero value if the PMT data has changed, since the last
     time this function was called.
      INPUT: void * - Pointer to the memory provided for the module. */
int          PMT_PSI_MODULE_PMT_changed(void *mem);

/* Adds an entry to the PMT, using the given values.  This function is
     used when there is no PMT data in the stream.
      INPUT: void * - Pointer to the memory provided for the module.
             u16    - Program number of the new entry. Same value used
                        in the MPEG2_PAT_MODULE_add_entry() function.
             u16    - PID of the PCR.
             u16    - PID of the video stream.
             u8     - One of the STREAM_TYPE_* values defined
                        in PMT_SECTION.h
             u16    - PID of the audio stream. 
             u8     - One of the STREAM_TYPE_* values defined
                        in PMT_SECTION.h */
void         PMT_PSI_MODULE_add_entry(void *mem,
                                      u16   program_number,
                                      u16   PCR_PID,
                                      u16   video_PID,
                                      u8    video_type,
                                      u16   audio_PID,
                                      u8    audio_type);

/* Returns a pointer to the PMT_SECTION for the current program.
      INPUT: void * - Pointer to the memory provided for the module.
      RETURN: The whole PMT_SECTION is stored - Pointer to a PMT_SECTION.
              No PMT_SECTION is stored        - NULL */
PMT_SECTION *PMT_PSI_MODULE_get_PMT(void *mem);

#ifdef __cplusplus
};
#endif

#endif
