/***************************************************************************
                          mpeg_ts.h  -  description
                             -------------------
   Functions for handling mpeg transport streams.
                             -------------------
    begin                : Sun Aug 8 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef _MPEG_TS_H_
#define _MPEG_TS_H_

#include "common/typedefs.h"
#include "common/Thread.h"
#include "common/File.h"
#include "TS_PKT.h"


#ifdef __cplusplus
extern "C" {
#endif

/* A player function to be used with the MPEG2_SYS_Play() function.
     This function can handle MPEG transport streams. */
void * THREAD_FUNC_TYPE MPEG_TS_Play(void *arg);

/* Finds the first program in the PAT.
      INPUT: fp - Points to the FILE with the MPEG transport stream. 
      RETURN: Program number is stored in the global struct 
                (sys.program_number).
              Success - A non-zero value
              Failure - 0 */
void MPEG_TS_Find_Program(FILE_PTR fp);

/* Creates a program, using the first audio and video streams
     found in the given MPEG transport stream.
      INPUT: fp - Points to the FILE with the MPEG transport stream.
      RETURN: Program number, of newly created program, is stored
                in the global struct (sys.program_number).
              Success - A non-zero value
              Failure - 0 */
void MPEG_TS_Create_Program(FILE_PTR fp);

/* Gets the stream ids and types from the transport stream.
      INPUT: fp   - Points to the FILE with the MPEG transport
                      stream.
             size - Number of elements in the PID and stream_type
                      arrays.
      OUTPUT: PID         - Array of stream ids found in the transport
                              stream.
              stream_type - Array of STREAM_TYPE_* values 
                              corresponding to the values in the
                              PID array.
      RETURN: The number of PIDs found. */
u32  MPEG_TS_get_PIDs(FILE_PTR fp, u16 *PID, u8 *stream_type, u32 size);

/* Tries to determine the type of data contained in the stream with
     the given PID.
      INPUT: fp  - Pointer to an open file that contains MPEG TS data.
             PID - PID of the stream to look for.
      RETURN: Success - One of the STREAM_TYPE_* values defined
                          in PMT_SECTION.h
              Failure - Zero. */
u8   MPEG_TS_stream_type(FILE_PTR fp, u16 PID);

/* Loads the PAT.
      INPUT: fp - Pointer to an open file that contains MPEG TS data. */
void MPEG_TS_load_PAT(FILE_PTR fp);

/* Loads the PMT.
      INPUT: fp - Pointer to an open file that contains MPEG TS data. */
void MPEG_TS_load_PMT(FILE_PTR fp);


/* Deals with the TS_PKTs decoded in MPEG_TS_decode().
      INPUT TS_PKT * - Pointer to the latest TS_PKT in the stream.
            void *   - Pointer to any private data needed by the
                         function.
      RETURN: Non-zero - Ready for the next TS_PKT.
              Zero     - Stop decoding TS_PKTs. */
typedef int (*MPEG_TS_decode_func)(TS_PKT *, void *);

/* Decodes the MPEG transport stream in the given file
      INPUT: fp      - Pointer to an open file that contains MPEG TS 
                         data.
             end_pos - Point at which the decoding will stop. Decoding
                         will stop at an earlier point, if an EOF is
                         encountered.
             func    - Function that will deal with the decoded 
                        TS_PKTs.
             arg     - Pointer to any private data for func. */
void MPEG_TS_decode(FILE_PTR            fp, 
                    FILE_POS            end_pos, 
                    MPEG_TS_decode_func func,
                    void               *arg);

#ifdef __cplusplus
};
#endif

#endif
