/***************************************************************************
                         mmx_dct32.c  -  description
                             -------------------
   MMX version of the 32pt DCT function. Suitable for all processors (PMMX, 
   P2, and later) that support the MMX instruction set.
                             -------------------
    begin                : Fri Apr 29 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


/* Does a 32pt DCT operation.
      INPUT: data - An array holding the 32 sub-band samples.
      OUTPUT: data - The 32 samples are replaced by 32 dct
                       coefficients. */
void dct32(AV1 *data)
{
#if defined(__GCC__)
/*
# define REVERSE(r1, r2) \
      "movq "       "%%mm"#r1",  %%mm"#r2"\n\t" \
      "psllq "      "$32, "     "%%mm"#r2"\n\t" \
      "punpckhwd "  "%%mm"#r2",  %%mm"#r1"\n\t" \
      "movq "       "%%mm"#r1",  %%mm"#r2"\n\t" \
      "psllq "      "$32, "     "%%mm"#r2"\n\t" \
      "punpckhwd "  "%%mm"#r2",  %%mm"#r1"\n\t"
      
# define BF_P(i, ls, rs) \
      "movq "     #i"(%1), "    "%%mm1\n\t" \
      "movq "     #i"(%0), "    "%%mm0\n\t" \
      REVERSE(1, 2) \
      "movq "       "(%2), "    "%%mm3\n\t" \
      "movq "       "%%mm0, "   "%%mm2\n\t" \
      "paddw "      "%%mm1, "   "%%mm0\n\t" \
      "psubw "      "%%mm1, "   "%%mm2\n\t" \
      "movq "       "%%mm2, "   "%%mm1\n\t" \
      "pmulhw "     "%%mm3, "   "%%mm2\n\t" \
      "pmullw "     "%%mm3, "   "%%mm1\n\t" \
      "psllw $"     #ls", "     "%%mm2\n\t" \
      "psrlw $"     #rs", "     "%%mm1\n\t" \
      "por "        "%%mm1, "   "%%mm2\n\t" \
      "movq "       "%%mm0, " #i"(%0)\n\t" \
      "movq "       "%%mm2, " #i"(%1)\n\t"
*/
    /* op_code       src         dest */
# define STAGE1_BF(ls, rs) \
      "movq "       "(%1), "    "%%mm1\n\t" \
      "movq "       "(%0), "    "%%mm0\n\t" \
      "movq "       "%%mm1, "   "%%mm2\n\t" \
      "psllq "      "$32, "     "%%mm2\n\t" \
      "movq "       "8(%0), "   "%%mm4\n\t" \
      "punpckhwd "  "%%mm2, "   "%%mm1\n\t" \
      "movq "       "%%mm1, "   "%%mm2\n\t" \
      "movq "       "-8(%1), "  "%%mm5\n\t" \
      "psllq "      "$32, "     "%%mm2\n\t" \
      "movq "       "(%2), "    "%%mm3\n\t" \
      "punpckhwd "  "%%mm2, "   "%%mm1\n\t" \
      "movq "       "%%mm5, "   "%%mm6\n\t" \
      "movq "       "%%mm0, "   "%%mm2\n\t" \
      "paddw "      "%%mm1, "   "%%mm0\n\t" \
      "psllq "      "$32, "     "%%mm6\n\t" \
      "psubw "      "%%mm1, "   "%%mm2\n\t" \
      "movq "       "8(%2), "   "%%mm7\n\t" \
      "movq "       "%%mm2, "   "%%mm1\n\t" \
      "pmulhw "     "%%mm3, "   "%%mm2\n\t" \
      "punpckhwd "  "%%mm6, "   "%%mm5\n\t" \
      "pmullw "     "%%mm3, "   "%%mm1\n\t" \
      "movq "       "%%mm5, "   "%%mm6\n\t" \
      "psllq "      "$32, "     "%%mm6\n\t" \
      "punpckhwd "  "%%mm6, "   "%%mm5\n\t" \
      "movq "       "%%mm4, "   "%%mm6\n\t" \
      "psllw $"     #ls", "     "%%mm2\n\t" \
      "paddw "      "%%mm5, "   "%%mm4\n\t" \
      "psubw "      "%%mm5, "   "%%mm6\n\t" \
      "psrlw $"     #rs", "     "%%mm1\n\t" \
      "movq "       "%%mm6, "   "%%mm5\n\t" \
      "pmulhw "     "%%mm7, "   "%%mm6\n\t" \
      "por "        "%%mm1, "   "%%mm2\n\t" \
      "pmullw "     "%%mm7, "   "%%mm5\n\t" \
      "psllw $"     #ls", "     "%%mm6\n\t" \
      "movq "       "%%mm0, "   "(%0)\n\t" \
      "psrlw $"     #rs", "     "%%mm5\n\t" \
      "movq "       "%%mm2, "   "(%1)\n\t" \
      "por "        "%%mm5, "   "%%mm6\n\t" \
      "movq "       "%%mm4, "   "8(%0)\n\t" \
      "movq "       "%%mm6, "   "-8(%1)\n\t"

    /* op_code       src         dest */
# define STAGE23_BF(inc, ls, rs) \
      "movq "       "(%1), "    "%%mm1\n\t" \
      "movq "       "(%0), "    "%%mm0\n\t" \
      "movq "       "%%mm1, "   "%%mm2\n\t" \
      "psllq "      "$32, "     "%%mm2\n\t" \
      "movq "   #inc"(%0), "    "%%mm4\n\t" \
      "punpckhwd "  "%%mm2, "   "%%mm1\n\t" \
      "movq "       "%%mm1, "   "%%mm2\n\t" \
      "movq "   #inc"(%1), "    "%%mm5\n\t" \
      "psllq "      "$32, "     "%%mm2\n\t" \
      "movq "       "(%2), "    "%%mm3\n\t" \
      "punpckhwd "  "%%mm2, "   "%%mm1\n\t" \
      "movq "       "%%mm4, "   "%%mm6\n\t" \
      "movq "       "%%mm0, "   "%%mm2\n\t" \
      "paddw "      "%%mm1, "   "%%mm0\n\t" \
      "psllq "      "$32, "     "%%mm6\n\t" \
      "psubw "      "%%mm1, "   "%%mm2\n\t" \
      "movq "       "%%mm2, "   "%%mm1\n\t" \
      "pmulhw "     "%%mm3, "   "%%mm2\n\t" \
      "punpckhwd "  "%%mm6, "   "%%mm4\n\t" \
      "pmullw "     "%%mm3, "   "%%mm1\n\t" \
      "movq "       "%%mm4, "   "%%mm6\n\t" \
      "psllw $"     #ls", "     "%%mm2\n\t" \
      "psrlw $"     #rs", "     "%%mm1\n\t" \
      "psllq "      "$32, "     "%%mm6\n\t" \
      "por "        "%%mm1, "   "%%mm2\n\t" \
      "punpckhwd "  "%%mm6, "   "%%mm4\n\t" \
      "movq "       "%%mm5, "   "%%mm6\n\t" \
      "paddw "      "%%mm4, "   "%%mm5\n\t" \
      "psubw "      "%%mm4, "   "%%mm6\n\t" \
      "movq "       "%%mm6, "   "%%mm4\n\t" \
      "pmulhw "     "%%mm3, "   "%%mm6\n\t" \
      "pmullw "     "%%mm3, "   "%%mm4\n\t" \
      "movq "       "%%mm0, "   "(%0)\n\t" \
      "psllw $"     #ls", "     "%%mm6\n\t" \
      "psrlw $"     #rs", "     "%%mm4\n\t" \
      "movq "       "%%mm2, "   "(%1)\n\t" \
      "por "        "%%mm4, "   "%%mm6\n\t" \
      "movq "       "%%mm5, "#inc"(%0)\n\t" \
      "movq "       "%%mm6, "#inc"(%1)\n\t"
      
    /* op_code       src         dest */
# define STAGE4_BF(ls, rs) \
      "movq "       "(%0), "    "%%mm0\n\t" \
      "pxor "       "%%mm2, "   "%%mm2\n\t" \
      "movq "       "%%mm0, "   "%%mm1\n\t" \
      "psubw "      "%%mm0, "   "%%mm2\n\t" \
      "movq "       "8(%0), "   "%%mm4\n\t" \
      "psllq "      "$32, "     "%%mm2\n\t" \
      "pxor "       "%%mm6, "   "%%mm6\n\t" \
      "punpckhwd "  "%%mm2, "   "%%mm1\n\t" \
      "movq "       "%%mm4, "   "%%mm5\n\t" \
      "movq "       "%%mm1, "   "%%mm2\n\t" \
      "psubw "      "%%mm4, "   "%%mm6\n\t" \
      "psllq "      "$32, "     "%%mm2\n\t" \
      "punpckhwd "  "%%mm2, "   "%%mm1\n\t" \
      "psllq "      "$32, "     "%%mm6\n\t" \
      "paddw "      "%%mm1, "   "%%mm0\n\t" \
      "punpckhwd "  "%%mm6, "   "%%mm5\n\t" \
      "movq "       "%%mm0, "   "%%mm1\n\t" \
      "pmullw "     "%%mm3, "   "%%mm0\n\t" \
      "movq "       "%%mm5, "   "%%mm6\n\t" \
      "pmulhw "     "%%mm3, "   "%%mm1\n\t" \
      "psllq "      "$32, "     "%%mm6\n\t" \
      "punpckhwd "  "%%mm6, "   "%%mm5\n\t" \
      "psrlw $"     #rs", "     "%%mm0\n\t" \
      "paddw "      "%%mm5, "   "%%mm4\n\t" \
      "psllw $"     #ls", "     "%%mm1\n\t" \
      "movq "       "%%mm4, "   "%%mm5\n\t" \
      "pmullw "     "%%mm3, "   "%%mm4\n\t" \
      "por "        "%%mm1, "   "%%mm0\n\t" \
      "pmulhw "     "%%mm3, "   "%%mm5\n\t" \
      "psrlw $"     #rs", "     "%%mm4\n\t" \
      "movq "       "%%mm0, "   "(%0)\n\t" \
      "psllw $"     #ls", "     "%%mm5\n\t" \
      "por "        "%%mm5, "   "%%mm4\n\t" \
      "movq "       "%%mm4, "   "8(%0)\n\t"

    /* op_code       src         dest */
# define STAGE5_BF(ls, rs) \
      "movq "       "(%0), "    "%%mm0\n\t" \
      "pxor "       "%%mm1, "   "%%mm1\n\t" \
      "movq "       "8(%0), "   "%%mm4\n\t" \
      "movq "       "%%mm0, "   "%%mm2\n\t" \
      "psubw "      "%%mm0, "   "%%mm1\n\t" \
      "psrld "      "$16, "     "%%mm2\n\t" \
      "pxor "       "%%mm5, "   "%%mm5\n\t" \
      "pslld "      "$16, "     "%%mm1\n\t" \
      "movq "       "%%mm4, "   "%%mm6\n\t" \
      "por "        "%%mm2, "   "%%mm1\n\t" \
      "psubw "      "%%mm4, "   "%%mm5\n\t" \
      "paddw "      "%%mm1, "   "%%mm0\n\t" \
      "psrld "      "$16, "     "%%mm6\n\t" \
      "movq "       "%%mm0, "   "%%mm1\n\t" \
      "pslld "      "$16, "     "%%mm5\n\t" \
      "pmullw "     "%%mm3, "   "%%mm0\n\t" \
      "por "        "%%mm6, "   "%%mm5\n\t" \
      "pmulhw "     "%%mm3, "   "%%mm1\n\t" \
      "paddw "      "%%mm5, "   "%%mm4\n\t" \
      "psrlw $"     #rs", "     "%%mm0\n\t" \
      "movq "       "%%mm4, "   "%%mm5\n\t" \
      "psllw $"     #ls", "     "%%mm1\n\t" \
      "pmullw "     "%%mm3, "   "%%mm4\n\t" \
      "por "        "%%mm1, "   "%%mm0\n\t" \
      "pmulhw "     "%%mm3, "   "%%mm5\n\t" \
      "movq "       "%%mm0, "   "%%mm1\n\t" \
      "psrlw $"     #rs", "     "%%mm4\n\t" \
      "pand "       "%%mm7, "   "%%mm1\n\t" \
      "psllw $"     #ls", "     "%%mm5\n\t" \
      "psrld "      "$16, "     "%%mm1\n\t" \
      "por "        "%%mm4, "   "%%mm5\n\t" \
      "pxor "       "%%mm6, "   "%%mm6\n\t" \
      "movq "       "%%mm5, "   "%%mm2\n\t" \
      "movq "       "%%mm5, "   "%%mm4\n\t" \
      "punpckhwd "  "%%mm6, "   "%%mm5\n\t" \
      "paddw "      "%%mm1, "   "%%mm0\n\t" \
      "paddw "      "%%mm5, "   "%%mm4\n\t" \
      "psrlq "      "$48, "     "%%mm2\n\t" \
      "movq "       "%%mm4, "   "%%mm6\n\t" \
      "punpcklwd "  "%%mm2, "   "%%mm2\n\t" \
      "psllq "      "$32, "     "%%mm6\n\t" \
      "paddw "      "%%mm2, "   "%%mm4\n\t" \
      "psrld "      "$16, "     "%%mm6\n\t" \
      "movq "       "%%mm0, "   "(%0)\n\t" \
      "paddw "      "%%mm6, "   "%%mm4\n\t" \
      "movq "       "%%mm4, "   "8(%0)\n\t"

    /* op_code     src         dest */
# define STAGE6 \
      "movq "     "(%0), "    "%%mm0\n\t" \
      "movl "     "%0, "      "%1\n\t" \
      "movq "     "8(%0), "   "%%mm1\n\t" \
      "movq "     "%%mm0, "   "%%mm3\n\t" \
      "movq "     "32(%0), "  "%%mm4\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "movq "     "40(%0), "  "%%mm5\n\t" \
      "psrld "    "$16, "     "%%mm3\n\t" \
      "movq "     "%%mm4, "   "%%mm7\n\t" \
      "paddw "    "%%mm1, "   "%%mm0\n\t" \
      "psrlq "    "$32, "     "%%mm2\n\t" \
      "movq "     "%%mm4, "   "%%mm6\n\t" \
      "psllq "    "$32, "     "%%mm3\n\t" \
      "paddw "    "%%mm2, "   "%%mm1\n\t" \
      "psrld "    "$16, "     "%%mm7\n\t" \
      "paddw "    "%%mm5, "   "%%mm4\n\t" \
      "psrlq "    "$32, "     "%%mm6\n\t" \
      "paddw "    "%%mm3, "   "%%mm1\n\t" \
      "psllq "    "$32, "     "%%mm7\n\t" \
      "movq "     "%%mm0, "   "(%0)\n\t" \
      "paddw "    "%%mm6, "   "%%mm5\n\t" \
      "movq "     "%%mm1, "   "8(%0)\n\t" \
      "paddw "    "%%mm7, "   "%%mm5\n\t" \
      "movq "     "%%mm4, "   "32(%0)\n\t" \
      "subl "     "$(8*2), "  "%1\n\t" \
      "movq "     "%%mm5, "   "40(%0)\n\t"
      
    /* op_code     src         dest */
# define SWAP_VALUES \
      "movq "     "(%1), "    "%%mm0\n\t" \
      "movq "     "32(%1), "  "%%mm1\n\t" \
      "movq "     "%%mm0, "   "%%mm4\n\t" \
      "punpcklwd ""%%mm1, "   "%%mm0\n\t" \
      "movq "     "16(%1), "  "%%mm2\n\t" \
      "punpckhwd ""%%mm1, "   "%%mm4\n\t" \
      "movq "     "48(%1), "  "%%mm3\n\t" \
      "movq "     "%%mm2, "   "%%mm5\n\t" \
      "punpcklwd ""%%mm3, "   "%%mm2\n\t" \
      "movq "     "%%mm0, "   "%%mm1\n\t" \
      "punpckhwd ""%%mm3, "   "%%mm5\n\t" \
      "movq "     "%%mm4, "   "%%mm6\n\t" \
      "punpckldq ""%%mm2, "   "%%mm0\n\t" \
      "movl "     "%1, "      "%0\n\t" \
      "punpckldq ""%%mm5, "   "%%mm4\n\t" \
      "movq "     "%%mm0, "   "(%0)\n\t" \
      "punpckhdq ""%%mm2, "   "%%mm1\n\t" \
      "movq "     "%%mm4, "   "16(%0)\n\t" \
      "punpckhdq ""%%mm5, "   "%%mm6\n\t" \
      "movq "     "%%mm1, "   "32(%0)\n\t" \
      "addl "     "$(4*2), "  "%1\n\t" \
      "movq "     "%%mm6, "   "48(%0)\n\t"
      
    /* op_code     src         dest */
# define ODD_VALUES \
      "movq "     "8(%1), "   "%%mm1\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "movq "     "16(%1), "  "%%mm4\n\t" \
      "movq "     "%%mm1, "   "%%mm3\n\t" \
      "psrld "    "$16, "     "%%mm2\n\t" \
      "movq "     "%%mm1, "   "%%mm6\n\t" \
      "psllq "    "$16, "     "%%mm3\n\t" \
      "movq "     "%%mm4, "   "%%mm7\n\t" \
      "psrlq "    "$16, "     "%%mm2\n\t" \
      "movl "     "%1, "      "%0\n\t" \
      "pslld "    "$16, "     "%%mm3\n\t" \
      "paddw "    "%%mm2, "   "%%mm0\n\t" \
      "psrld "    "$16, "     "%%mm6\n\t" \
      "addl "     "$(8*2), "  "%1\n\t" \
      "psllq "    "$16, "     "%%mm7\n\t" \
      "paddw "    "%%mm3, "   "%%mm0\n\t" \
      "psrlq "    "$16, "     "%%mm6\n\t" \
      "movq "     "%%mm0, "   "%%mm5\n\t" \
      "pslld "    "$16, "     "%%mm7\n\t" \
      "paddw "    "%%mm6, "   "%%mm1\n\t" \
      "movq "     "%%mm5, "   "(%0)\n\t" \
      "paddw "    "%%mm7, "   "%%mm1\n\t" \
      "movq "     "%%mm4, "   "%%mm0\n\t" \
      "movq "     "%%mm1, "   "8(%0)\n\t"

    /* op_code     src         dest */
# define ODD_VALUES_END \
      "movq "     "8(%1), "   "%%mm1\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "movq "     "%%mm1, "   "%%mm3\n\t" \
      "psrld "    "$16, "     "%%mm2\n\t" \
      "psllq "    "$16, "     "%%mm3\n\t" \
      "movq "     "%%mm1, "   "%%mm5\n\t" \
      "psrlq "    "$16, "     "%%mm2\n\t" \
      "pslld "    "$16, "     "%%mm3\n\t" \
      "paddw "    "%%mm2, "   "%%mm0\n\t" \
      "psrld "    "$16, "     "%%mm5\n\t" \
      "paddw "    "%%mm3, "   "%%mm0\n\t" \
      "psrlq "    "$16, "     "%%mm5\n\t" \
      "paddw "    "%%mm5, "   "%%mm1\n\t" \
      "movq "     "%%mm0, "   "(%1)\n\t" \
      "movq "     "%%mm1, "   "8(%1)\n\t"

   __asm__ __volatile__(
    /* data order:  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15
                   16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 */
    /* STAGE 1: top=data[ 0], bot=data[28], cos_val=cos_const[0] 
         registers used: %0 = top
                         %1 = bot
                         %2 = cos_val */
    /* op_code     src         dest */
      STAGE1_BF(5, 11)
      "addl "     "$(8*2), "  "%0\n\t" /* top += 8; */
      "subl "     "$(8*2), "  "%1\n\t" /* bot -= 8; */
      "addl "     "$(8*2), "  "%2\n\t" /* cos_val += 8; */
      STAGE1_BF(5, 11)
      "subl "     "$(8*2), "  "%0\n\t" /* top -= 8; */
      "subl "     "$(8*2), "  "%1\n\t" /* bot -= 8; */
      "addl "     "$(8*2), "  "%2\n\t" /* cos_val += 8; */
    
    /* data order:  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15
                   19-18-17-16 23-22-21-20 27-26-25-24 31-30-29-28 */
    /* STAGE 2: top=data[ 0], bot=data[12], cos_val=cos_const[16] 
         registers used: %0 = top
                         %1 = bot
                         %2 = cos_val */
    /* op_code     src         dest */
      STAGE23_BF(32, 4, 12)
      "addl "     "$(4*2), "  "%0\n\t" /* top += 4; */
      "subl "     "$(4*2), "  "%1\n\t" /* bot -= 4; */
      "addl "     "$(4*2), "  "%2\n\t" /* cos_val += 4; */
      STAGE23_BF(32, 4, 12)
      "subl "     "$(4*2), "  "%0\n\t" /* top -= 4; */
      "subl "     "$(4*2), "  "%1\n\t" /* bot -= 4; */
      "addl "     "$(4*2), "  "%2\n\t" /* cos_val += 4; */

    /* data order:  0  1  2  3  4  5  6  7 11-10--9--8 15-14-13-12
                   16 17 18 19 20 21 22 23 27-26-25-24 31-30-29-28 */
    /* STAGE 3: top=data[ 0], bot=data[ 4], cos_val=cos_const[16+8] 
         registers used: %0 = top
                         %1 = bot
                         %2 = cos_val */
    /* op_code     src         dest */
      STAGE23_BF(16, 4, 12)
      "addl "     "$(16*2), " "%0\n\t" /* top += 16; */
      "addl "     "$(16*2), " "%1\n\t" /* bot += 16; */
      STAGE23_BF(16, 4, 12)
      "addl "     "$(4*2), "  "%2\n\t" /* cos_val += 4; */
      "subl "     "$(16*2), " "%0\n\t" /* top -= 16; */
    
    /* data order:  0  1  2  3  7--6--5--4  8  9 10 11 15-14-13-12
                   16 17 18 19 23-22-21-20 24 25 26 27 31-30-29-28 */
    /* STAGE 4: top=data[ 0], bot=not used, cos_val=cos_const[16+8+4] 
         registers used: %0   = top
                         %2   = cos_val
                         %4   = loop counter
                        %%mm3 = STAGE4 cos constants */
    /* op_code     src         dest */
      "movq "     "(%2), "    "%%mm3\n\t"
      "STAGE4_LOOP:\n\t"
      STAGE4_BF(4, 12)
      "addl "     "$(8*2), "  "%0\n\t" /* top += 8; */
      "dec "                  "%4\n\t"
      "jnz "      "STAGE4_LOOP\n\t"
      "addl "     "$(4*2), "  "%2\n\t" /* cos_val += 4; */
      "subl "     "$(32*2), " "%0\n\t" /* top -= 32; */

    /* data order:  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15
                   16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 */
    /* STAGE 5: top=data[ 0], bot=not used, cos_val=cos_const[16+8+4+4] 
         registers used: %0   = top
                         %2   = cos_val
                         %3   = mask
                         %4   = loop counter
                        %%mm3 = STAGE5 cos constants */
    /* op_code     src         dest */
      "movd "     "%3, "      "%%mm7\n\t"
      "movq "     "(%2), "    "%%mm3\n\t"
      "psllq "    "$32, "     "%%mm7\n\t"
      "addl "     "$4, "      "%4\n\t"
      "STAGE5_LOOP:\n\t"
      STAGE5_BF(4, 12)
      "addl "     "$(8*2), "  "%0\n\t" /* top += 8; */
      "dec "                  "%4\n\t"
      "jnz "      "STAGE5_LOOP\n\t"
      "subl "     "$(24*2), " "%0\n\t" /* top -= 24; */

    /* data order:  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15
                   16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 */
    /* STAGE 6: top=data[ 8], bot=not used, cos_val=not used
         registers used: %0 = top */
      STAGE6

    /* data order:  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15
                   16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 */
    /* SWAP VALUES: top=data[ 8], bot=data[ 0], cos_val=not used
         registers used: %0 = top 
                         %1 = bot
                         %4 = loop counter */
    /* op_code     src         dest */
      "addl "     "$2, "      "%4\n\t"
      "SWAP_LOOP:\n\t"
      SWAP_VALUES
      "dec "                  "%4\n\t"
      "jnz "      "SWAP_LOOP\n\t"
      "subl "     "$(8*2), "  "%1\n\t"

    /* data order:  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15
                   16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 */
    /* ODD VALUES: top=data[ 4], bot=data[ 0], cos_val=not used
         registers used: %0 = top 
                         %1 = bot
                         %4 = loop counter */
    /* op_code     src         dest */
      "addl "     "$3, "      "%4\n\t"
      "movq "     "(%1), "    "%%mm0\n\t"
      "ODD_LOOP:\n\t"
      ODD_VALUES
      "dec "                  "%4\n\t"
      "jnz "      "ODD_LOOP\n\t"
      ODD_VALUES_END

   /* output list */
   :

   /* input list */
   : "q" (data),
     "q" (data+28),
     "q" (cos_const),
     "r" (0xFFFFFFFF),
     "r" (4)
   );

#elif defined(__MSVC__)
        /* op_code   dest     src */
# define STAGE1_BF(ls, rs) \
     __asm movq      mm1,     [esi] \
     __asm movq      mm0,     [edi] \
     __asm movq      mm2,     mm1 \
     __asm psllq     mm2,     32 \
     __asm movq      mm4,     [edi+8] \
     __asm punpckhwd mm1,     mm2 \
     __asm movq      mm2,     mm1 \
     __asm movq      mm5,     [esi-8] \
     __asm psllq     mm2,     32 \
     __asm movq      mm3,     [edx] \
     __asm punpckhwd mm1,     mm2 \
     __asm movq      mm6,     mm5 \
     __asm movq      mm2,     mm0 \
     __asm paddw     mm0,     mm1 \
     __asm psllq     mm6,     32 \
     __asm psubw     mm2,     mm1 \
     __asm movq      mm7,     [edx+8] \
     __asm movq      mm1,     mm2 \
     __asm pmulhw    mm2,     mm3 \
     __asm punpckhwd mm5,     mm6 \
     __asm pmullw    mm1,     mm3 \
     __asm movq      mm6,     mm5 \
     __asm psllq     mm6,     32 \
     __asm punpckhwd mm5,     mm6 \
     __asm movq      mm6,     mm4 \
     __asm psllw     mm2,     ls \
     __asm paddw     mm4,     mm5 \
     __asm psubw     mm6,     mm5 \
     __asm psrlw     mm1,     rs \
     __asm movq      mm5,     mm6 \
     __asm pmulhw    mm6,     mm7 \
     __asm por       mm2,     mm1 \
     __asm pmullw    mm5,     mm7 \
     __asm psllw     mm6,     ls \
     __asm movq      [edi],   mm0 \
     __asm psrlw     mm5,     rs \
     __asm movq      [esi],   mm2 \
     __asm por       mm6,     mm5 \
     __asm movq      [edi+8], mm4 \
     __asm movq      [esi-8], mm6

        /* op_code   dest     src */
# define STAGE23_BF(inc, ls, rs) \
     __asm movq      mm1,     [esi] \
     __asm movq      mm0,     [edi] \
     __asm movq      mm2,     mm1 \
     __asm psllq     mm2,     32 \
     __asm movq      mm4,     [edi+inc] \
     __asm punpckhwd mm1,     mm2 \
     __asm movq      mm2,     mm1 \
     __asm movq      mm5,     [esi+inc] \
     __asm psllq     mm2,     32 \
     __asm movq      mm3,     [edx] \
     __asm punpckhwd mm1,     mm2 \
     __asm movq      mm6,     mm4 \
     __asm movq      mm2,     mm0 \
     __asm paddw     mm0,     mm1 \
     __asm psllq     mm6,     32 \
     __asm psubw     mm2,     mm1 \
     __asm movq      mm1,     mm2 \
     __asm pmulhw    mm2,     mm3 \
     __asm punpckhwd mm4,     mm6 \
     __asm pmullw    mm1,     mm3 \
     __asm movq      mm6,     mm4 \
     __asm psllw     mm2,     ls \
     __asm psrlw     mm1,     rs \
     __asm psllq     mm6,     32 \
     __asm por       mm2,     mm1 \
     __asm punpckhwd mm4,     mm6 \
     __asm movq      mm6,     mm5 \
     __asm paddw     mm5,     mm4 \
     __asm psubw     mm6,     mm4 \
     __asm movq      mm4,     mm6 \
     __asm pmulhw    mm6,     mm3 \
     __asm pmullw    mm4,     mm3 \
     __asm movq      [edi],   mm0 \
     __asm psllw     mm6,     ls \
     __asm psrlw     mm4,     rs \
     __asm movq      [esi],   mm2 \
     __asm por       mm6,     mm4 \
     __asm movq    [edi+inc], mm5 \
     __asm movq    [esi+inc], mm6

        /* op_code   dest     src */
# define STAGE4_BF(ls, rs) \
     __asm movq      mm0,     [edi] \
     __asm pxor      mm2,     mm2 \
     __asm movq      mm1,     mm0 \
     __asm psubw     mm2,     mm0 \
     __asm movq      mm4,     [edi+8] \
     __asm psllq     mm2,     32 \
     __asm pxor      mm6,     mm6 \
     __asm punpckhwd mm1,     mm2 \
     __asm movq      mm5,     mm4 \
     __asm movq      mm2,     mm1 \
     __asm psubw     mm6,     mm4 \
     __asm psllq     mm2,     32 \
     __asm punpckhwd mm1,     mm2 \
     __asm psllq     mm6,     32 \
     __asm paddw     mm0,     mm1 \
     __asm punpckhwd mm5,     mm6 \
     __asm movq      mm1,     mm0 \
     __asm pmullw    mm0,     mm3 \
     __asm movq      mm6,     mm5 \
     __asm pmulhw    mm1,     mm3 \
     __asm psllq     mm6,     32 \
     __asm punpckhwd mm5,     mm6 \
     __asm psrlw     mm0,     rs \
     __asm paddw     mm4,     mm5 \
     __asm psllw     mm1,     ls \
     __asm movq      mm5,     mm4 \
     __asm pmullw    mm4,     mm3 \
     __asm por       mm0,     mm1 \
     __asm pmulhw    mm5,     mm3 \
     __asm psrlw     mm4,     rs \
     __asm movq      [edi],   mm0 \
     __asm psllw     mm5,     ls \
     __asm por       mm4,     mm5 \
     __asm movq      [edi+8], mm4

        /* op_code   dest     src */
# define STAGE5_BF(ls, rs) \
     __asm movq      mm0,     [edi] \
     __asm pxor      mm1,     mm1 \
     __asm movq      mm4,     [edi+8] \
     __asm movq      mm2,     mm0 \
     __asm psubw     mm1,     mm0 \
     __asm psrld     mm2,     16 \
     __asm pxor      mm5,     mm5 \
     __asm pslld     mm1,     16 \
     __asm movq      mm6,     mm4 \
     __asm por       mm1,     mm2 \
     __asm psubw     mm5,     mm4 \
     __asm paddw     mm0,     mm1 \
     __asm psrld     mm6,     16 \
     __asm movq      mm1,     mm0 \
     __asm pslld     mm5,     16 \
     __asm pmullw    mm0,     mm3 \
     __asm por       mm5,     mm6 \
     __asm pmulhw    mm1,     mm3 \
     __asm paddw     mm4,     mm5 \
     __asm psrlw     mm0,     rs \
     __asm movq      mm5,     mm4 \
     __asm psllw     mm1,     ls \
     __asm pmullw    mm4,     mm3 \
     __asm por       mm0,     mm1 \
     __asm pmulhw    mm5,     mm3 \
     __asm movq      mm1,     mm0 \
     __asm psrlw     mm4,     rs \
     __asm pand      mm1,     mm7 \
     __asm psllw     mm5,     ls \
     __asm psrld     mm1,     16 \
     __asm por       mm5,     mm4 \
     __asm pxor      mm6,     mm6 \
     __asm movq      mm2,     mm5 \
     __asm movq      mm4,     mm5 \
     __asm punpckhwd mm5,     mm6 \
     __asm paddw     mm0,     mm1 \
     __asm paddw     mm4,     mm5 \
     __asm psrlq     mm2,     48 \
     __asm movq      mm6,     mm4 \
     __asm punpcklwd mm2,     mm2 \
     __asm psllq     mm6,     32 \
     __asm paddw     mm4,     mm2 \
     __asm psrld     mm6,     16 \
     __asm movq      [edi],   mm0 \
     __asm paddw     mm4,     mm6 \
     __asm movq      [edi+8], mm4

        /* op_code   dest     src */
# define STAGE6 \
     __asm movq      mm0,     [edi] \
     __asm mov       esi,     edi \
     __asm movq      mm1,     [edi+8] \
     __asm movq      mm3,     mm0 \
     __asm movq      mm4,     [edi+32] \
     __asm movq      mm2,     mm0 \
     __asm movq      mm5,     [edi+40] \
     __asm psrld     mm3,     16 \
     __asm movq      mm7,     mm4 \
     __asm paddw     mm0,     mm1 \
     __asm psrlq     mm2,     32 \
     __asm movq      mm6,     mm4 \
     __asm psllq     mm3,     32 \
     __asm paddw     mm1,     mm2 \
     __asm psrld     mm7,     16 \
     __asm paddw     mm4,     mm5 \
     __asm psrlq     mm6,     32 \
     __asm paddw     mm1,     mm3 \
     __asm psllq     mm7,     32 \
     __asm movq      [edi],   mm0 \
     __asm paddw     mm5,     mm6 \
     __asm movq      [edi+8], mm1 \
     __asm paddw     mm5,     mm7 \
     __asm movq      [edi+32],mm4 \
     __asm sub       esi,     (8*2) \
     __asm movq      [edi+40],mm5

        /* op_code   dest     src */
# define SWAP_VALUES \
     __asm movq      mm0,     [esi] \
     __asm movq      mm1,     [esi+32] \
     __asm movq      mm4,     mm0 \
     __asm punpcklwd mm0,     mm1 \
     __asm movq      mm2,     [esi+16] \
     __asm punpckhwd mm4,     mm1 \
     __asm movq      mm3,     [esi+48] \
     __asm movq      mm5,     mm2 \
     __asm punpcklwd mm2,     mm3 \
     __asm movq      mm1,     mm0 \
     __asm punpckhwd mm5,     mm3 \
     __asm movq      mm6,     mm4 \
     __asm punpckldq mm0,     mm2 \
     __asm mov       edi,     esi \
     __asm punpckldq mm4,     mm5 \
     __asm movq      [edi],   mm0 \
     __asm punpckhdq mm1,     mm2 \
     __asm movq      [edi+16],mm4 \
     __asm punpckhdq mm6,     mm5 \
     __asm movq      [edi+32],mm1 \
     __asm add       esi,     (4*2) \
     __asm movq      [edi+48],mm6

        /* op_code   dest     src */
# define ODD_VALUES \
     __asm movq      mm1,     [esi+8] \
     __asm movq      mm2,     mm0 \
     __asm movq      mm4,     [esi+16] \
     __asm movq      mm3,     mm1 \
     __asm psrld     mm2,     16 \
     __asm movq      mm6,     mm1 \
     __asm psllq     mm3,     16 \
     __asm movq      mm7,     mm4 \
     __asm psrlq     mm2,     16 \
     __asm mov       edi,     esi \
     __asm pslld     mm3,     16 \
     __asm paddw     mm0,     mm2 \
     __asm psrld     mm6,     16 \
     __asm add       esi,     (8*2) \
     __asm psllq     mm7,     16 \
     __asm paddw     mm0,     mm3 \
     __asm psrlq     mm6,     16 \
     __asm movq      mm5,     mm0 \
     __asm pslld     mm7,     16 \
     __asm paddw     mm1,     mm6 \
     __asm movq      [edi],   mm5 \
     __asm paddw     mm1,     mm7 \
     __asm movq      mm0,     mm4 \
     __asm movq      [edi+8], mm1

        /* op_code   dest     src */
# define ODD_VALUES_END \
     __asm movq      mm1,     [esi+8] \
     __asm movq      mm2,     mm0 \
     __asm movq      mm3,     mm1 \
     __asm psrld     mm2,     16 \
     __asm psllq     mm3,     16 \
     __asm movq      mm5,     mm1 \
     __asm psrlq     mm2,     16 \
     __asm pslld     mm3,     16 \
     __asm paddw     mm0,     mm2 \
     __asm psrld     mm5,     16 \
     __asm paddw     mm0,     mm3 \
     __asm psrlq     mm5,     16 \
     __asm paddw     mm1,     mm5 \
     __asm movq      [esi],   mm0 \
     __asm movq      [esi+8], mm1

   /* op_code     dest     src */
   __asm {
      mov         edi,     dword ptr[data]
      mov         edx,     offset cos_const
      mov         esi,     edi
      mov         ecx,     0xFFFFFFFF
      add         esi,     (28*2)
      mov         ebx,     4

   /* data order:  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15
                  16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 */
   /* STAGE 1: top=data[ 0], bot=data[28], cos_val=cos_const[0] 
        registers used: edi = top
                        esi = bot
                        edx = cos_val */
   /* op_code     dest     src */
      STAGE1_BF(5, 11)
      add         edi,     (8*2)    /* top += 8; */
      sub         esi,     (8*2)    /* bot -= 8; */
      add         edx,     (8*2)    /* cos_val += 8; */
      STAGE1_BF(5, 11)
      sub         edi,     (8*2)    /* top -= 8; */
      sub         esi,     (8*2)    /* bot -= 8; */
      add         edx,     (8*2)    /* cos_val += 8; */

    /* data order:  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15
                   19-18-17-16 23-22-21-20 27-26-25-24 31-30-29-28 */
    /* STAGE 2: top=data[ 0], bot=data[12], cos_val=cos_const[16] 
         registers used: edi = top
                         esi = bot
                         edx = cos_val */
   /* op_code     dest     src */
      STAGE23_BF(32, 4, 12)
      add         edi,     (4*2)    /* top += 4; */
      sub         esi,     (4*2)    /* bot -= 4; */
      add         edx,     (4*2)    /* cos_val += 4; */
      STAGE23_BF(32, 4, 12)
      sub         edi,     (4*2)    /* top -= 4; */
      sub         esi,     (4*2)    /* bot -= 4; */
      add         edx,     (4*2)    /* cos_val += 4; */

    /* data order:  0  1  2  3  4  5  6  7 11-10--9--8 15-14-13-12
                   16 17 18 19 20 21 22 23 27-26-25-24 31-30-29-28 */
    /* STAGE 3: top=data[ 0], bot=data[ 4], cos_val=cos_const[16+8] 
         registers used: edi = top
                         esi = bot
                         edx = cos_val */
   /* op_code     dest     src */
      STAGE23_BF(16, 4, 12)
      add         edi,     (16*2)   /* top += 16; */
      add         esi,     (16*2)   /* bot += 16; */
      STAGE23_BF(16, 4, 12)
      add         edx,     (4*2)    /* cos_val += 4; */
      sub         edi,     (16*2)   /* top -= 16; */
    
    /* data order:  0  1  2  3  7--6--5--4  8  9 10 11 15-14-13-12
                   16 17 18 19 23-22-21-20 24 25 26 27 31-30-29-28 */
    /* STAGE 4: top=data[ 0], bot=not used, cos_val=cos_const[16+8+4] 
         registers used: edi = top
                         edx = cos_val
                         ebx = loop counter
                         mm3 = STAGE4 cos constants */
   /* op_code     dest     src */
      movq        mm3,     [edx]
   STAGE4_LOOP:
      STAGE4_BF(4, 12)
      add         edi,     (8*2)    /* top += 8; */
      dec         ebx
      jnz         STAGE4_LOOP
      add         edx,     (4*2)    /* cos_val += 4; */
      sub         edi,     (32*2)   /* top -= 32; */

    /* data order:  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15
                   16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 */
    /* STAGE 5: top=data[ 0], bot=not used, cos_val=cos_const[16+8+4+4] 
         registers used: edi = top
                         edx = cos_val
                         ecx = mask
                         ebx = loop counter
                         mm3 = STAGE5 cos constants */
   /* op_code     dest     src */
      movd        mm7,     ecx
      movq        mm3,     [edx]
      psllq       mm7,     32
      add         ebx,     4
   STAGE5_LOOP:
      STAGE5_BF(4, 12)
      add         edi,     (8*2)    /* top += 8; */
      dec         ebx
      jnz         STAGE5_LOOP
      sub         edi,     (24*2)   /* top -= 24; */

    /* data order:  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15
                   16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 */
    /* STAGE 6: top=data[ 8], bot=not used, cos_val=not used
         registers used: edi = top */
      STAGE6

    /* data order:  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15
                   16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 */
    /* SWAP VALUES: top=data[ 8], bot=data[ 0], cos_val=not used
         registers used: edi = top 
                         esi = bot
                         ebx = loop counter */
   /* op_code     dest     src */
      add         ebx,     2
   SWAP_LOOP:
      SWAP_VALUES
      dec         ebx
      jnz         SWAP_LOOP
      sub         esi,     (8*2)

    /* data order:  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15
                   16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 */
    /* ODD VALUES: top=data[ 4], bot=data[ 0], cos_val=not used
         registers used: edi = top 
                         esi = bot
                         ebx = loop counter */
   /* op_code     dest     src */
      add         ebx,     3
      movq        mm0,     [esi]
   ODD_LOOP:
      ODD_VALUES
      dec         ebx
      jnz         ODD_LOOP
      ODD_VALUES_END
   }

#else
# error "Don't know how to insert inline ASM code for this compiler"
#endif
}

