/***************************************************************************
                         inv_trans.h  -  description
                             -------------------
    begin                : Fri Jul 28 2006
    copyright            : (C) 2006 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __INV_TRANS_H__
#define __INV_TRANS_H__

#include "common/typedefs.h"


#ifdef __cplusplus
extern "C" {
#endif

typedef void (*VC1_INV_TRANS_FUNC)(s16 *);

extern const VC1_INV_TRANS_FUNC VC1_inv_trans[];

/* Does the VC1 8x8 inverse transformation.
      INPUT: data - An array holding the 64 coefficients.
      OUTPUT: data - The 64 coefficients are replaced by 64
                       recovered samples. */
void VC1_8x8_inv_trans(s16 *data);

/* Does 2 VC1 8x4 inverse transformations.
      INPUT: data - An array holding the 64 coefficients.
      OUTPUT: data - The 64 coefficients are replaced by 64
                       recovered samples. */
void VC1_8x4_inv_trans(s16 *data);

/* Does 2 VC1 4x8 inverse transformations.
      INPUT: data - An array holding the 64 coefficients.
      OUTPUT: data - The 64 coefficients are replaced by 64
                       recovered samples. */
void VC1_4x8_inv_trans(s16 *data);

/* Does 4 VC1 4x4 inverse transformation.
      INPUT: data - An array holding the 64 coefficients.
      OUTPUT: data - The 64 coefficients are replaced by 64
                       recovered samples. */
void VC1_4x4_inv_trans(s16 *data);


/*#define USE_PRECISE_OVERLAP*/

/* Does the overlap filtering operation in the vertical direction. */
#ifndef USE_PRECISE_OVERLAP
void VC1_overlap_Y(s16 *blk, u8 *outp, int inc);
#else
void VC1_overlap_Y(s16 *blk, s16 *pre, s16 *preB, s16 *preT, u8 *outp, int inc);
#endif

/* Does the overlap filtering operation in the horizontal direction. */
#ifndef USE_PRECISE_OVERLAP
void VC1_overlap_X(s16 *blk, u8 *outp, int inc);
void VC1_overlap_X_last(u8 *outp, int inc);
#else
void VC1_overlap_X(s16 *blk, s16 *pre, s16 *pre2, u8 *outp, int inc);
void VC1_overlap_X_last(s16 *blk, s16 *pre, u8 *outp, int inc);
#endif

#ifdef __cplusplus
};
#endif

#endif

