/***************************************************************************
                          winmain.c  -  description
                             -------------------
   Handle all the win32 stuff.
                             -------------------
    begin                : Thurs Feb 3 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdlib.h>
#include "winmain.h"
#include <commctrl.h>

#include "resource.h"
#include "FileDLG.h"
#include "file_actions.h"
#include "control_actions.h"
#include "commands.h"
#include "MPEG2_SYS.h"

/* #define __USE_CLIPPER__ 1 */

const char         *ERRMSG_TITLE   = "ERROR [mav_player]";
HINSTANCE           app            = NULL;
HWND                win            = NULL;
LPDIRECTDRAW        lpDD           = NULL;
LPDIRECTDRAWSURFACE surface        = NULL,
                    overlay        = NULL,
                    overlay_back   = NULL;
u32                 overlay_colour = 0,
                    win_mode       = WIN_MODE_NORMAL,
                    screen_wid     = 0,
                    screen_hgt     = 0,
                    win_wid        = 0,
                    win_hgt        = 0;
int                 win_x          = 0,
                    win_y          = 0;


static const char *CLASSNAME            = "MAV_PLAYER",
                  *win_title            = "MPEG Audio/Video Player";
static COLORREF    overlay_colour_24bit = RGB(255, 0, 255);
static u32         last_command         = CMD_PLAY;
static HMENU       win_menu             = NULL;
static RECT        last_pos;

#ifdef __USE_CLIPPER__
static LPDIRECTDRAWCLIPPER clipper = NULL;
#endif

static LRESULT CALLBACK event_handler(HWND hw, UINT msg, 
                                      WPARAM p1, LPARAM p2);
static int  init_DirectDraw();
static void free_DirectDraw();
static void get_overlay_colour();
static void PaintOverlay();
static void RestoreWindow();
static void MaximizeWindow();


int WINAPI WinMain(HINSTANCE hInstance,      /* Instance */
                   HINSTANCE hPrevInstance,  /* Previous Instance */
                   LPSTR     lpCmdLine,      /* Command Line Parameters */
                   int       nCmdShow)       /* Window Show State */
{
   WNDCLASS    wc;
   RECT        box;
   MSG         msg;
   HACCEL      acc;
   const char *err_msg = NULL;

   msg.wParam = 0;
   app        = hInstance;

   InitCommonControls();   /* load common controls library */

   /* allocate memory */
   if (!MPEG2_SYS_alloc(&sys)) {
      err_msg = "Cannot allocate memory for MPEG system.";
      goto LEAVE_WINMAIN;
   }

   /* Register Window Class */
   wc.style         = CS_VREDRAW | CS_HREDRAW | CS_OWNDC;
   wc.lpfnWndProc   = (WNDPROC)event_handler;
   wc.cbClsExtra    = 0;
   wc.cbWndExtra    = 0;
   wc.hInstance     = hInstance;
   wc.hIcon         = LoadIcon(NULL, IDI_WINLOGO);
   wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
   wc.hbrBackground = (HBRUSH)COLOR_WINDOW;
   wc.lpszMenuName  = NULL; /*MAKEINTRESOURCE(IDR_MAIN_MENU);*/
   wc.lpszClassName = CLASSNAME;

   if (!RegisterClass(&wc)) { /* Attempt To Register The Window Class */
      err_msg = "Failed to register the window class.";
      goto LEAVE_WINMAIN;
   }

   /* get the screen size */
   if (!GetWindowRect(GetDesktopWindow(), &box)) {
      err_msg = "Cannot get the screen size.";
      goto LEAVE_WINMAIN;
   }
   screen_wid = (u32)box.right;
   screen_hgt = (u32)box.bottom;

   /* get the initial window size */
   box.left   =   0;
   box.top    =   0;
   box.right  = 320;
   box.bottom = 240;
   AdjustWindowRectEx(&box, DWSTYLE, MENU_PRESENT, DWEXSTYLE);

   /* Create Window */
   win = CreateWindowEx(DWEXSTYLE,
                        CLASSNAME, 
                        win_title,
                        DWSTYLE,
                        0, 0,
                        (box.right - box.left),
                        (box.bottom - box.top),
                        NULL,
                        NULL,
                        hInstance,
                        NULL);
   if (!win) {
      err_msg = "Failed to create a window.";
      goto LEAVE_WINMAIN;
   }
   win_menu = LoadMenu(hInstance, MAKEINTRESOURCE(IDR_MAIN_MENU));
   SetMenu(win, win_menu);
   
   FileDLG_init(win);   /* initialize File Dialog */

   /* start up direct draw */
   if (!init_DirectDraw())
      goto LEAVE_WINMAIN;
   do {
      get_overlay_colour();
   } while (!overlay_colour);

   /* put the window on the screen */
   acc = LoadAccelerators(hInstance, MAKEINTRESOURCE(IDR_ACCELERATOR1));
   ShowWindow(win, SW_SHOW);
   SetForegroundWindow(win);
   SetFocus(win);
   GetWindowRect(win, &last_pos);

   while (GetMessage(&msg, NULL, 0, 0)) {
      if (!TranslateAccelerator(win, acc, &msg))
         DispatchMessage(&msg);
   }

LEAVE_WINMAIN:
   /* Shutdown */
   sys.data.command = CMD_STOP;
   MPEG2_SYS_free(&sys);
   free_DirectDraw();

   if (win) {
      if (win_menu) {
         SetMenu(win, NULL);
         DestroyMenu(win_menu);
      }
      DestroyWindow(win);
   }

   UnregisterClass(CLASSNAME, hInstance);

   if (err_msg) {
      MessageBox(NULL, err_msg,
                 ERRMSG_TITLE, (MB_OK|MB_ICONEXCLAMATION));
   }
   return msg.wParam;
}

void resize_DirectDrawOverlay()
{
   RECT        dest;
   DDOVERLAYFX fx;

   if (!GetClientRect(win, &dest))
      return;

   ClientToScreen(win, (POINT *) &(dest.left));
   ClientToScreen(win, (POINT *) &(dest.right));
   dest.left   += (long)win_x;
   dest.top    += (long)win_y;
   dest.right   = dest.left + (long)win_wid;
   dest.bottom  = dest.top  + (long)win_hgt;

   memset(&fx, 0, sizeof(DDOVERLAYFX));
   fx.dwSize = sizeof(DDOVERLAYFX);
   fx.dckDestColorkey.dwColorSpaceHighValue = overlay_colour;
   fx.dckDestColorkey.dwColorSpaceLowValue  = overlay_colour;

   IDirectDrawSurface_UpdateOverlay(overlay, NULL, surface, &dest,
                                    (DDOVER_SHOW | 
                                     DDOVER_KEYDESTOVERRIDE),
                                    &fx);
   InvalidateRect(win, NULL, FALSE);
}

void get_win_size(u32 disp_wid, u32 disp_hgt)
{
   u32 new_wid, new_hgt;

   if (win_mode == WIN_MODE_MAXIMIZED) {
      new_wid = screen_wid;
      new_hgt = screen_hgt;
   }
   else {
      new_wid = (u32)GetSystemMetrics(SM_CXFULLSCREEN) - 8;
      new_hgt = (u32)GetSystemMetrics(SM_CYFULLSCREEN) - 8;
   }

   if (disp_wid > new_wid)
      disp_wid = new_wid;
   if (disp_hgt > new_hgt)
      disp_hgt = new_hgt;

   new_wid  = disp_hgt; new_hgt  = disp_wid;
   new_wid *= win_wid;  new_hgt *= win_hgt;
   new_wid /= win_hgt;  new_hgt /= win_wid;
   
   if (new_wid >= disp_wid) {
      win_wid = disp_wid;
      win_x   = 0;
   }
   else {
      win_wid = new_wid;
      win_x   = (int)(disp_wid - new_wid) / 2;
      if (!win_x)
         win_x = 1;
   }
   
   if (new_hgt >= disp_hgt) {
      win_hgt = disp_hgt;
      win_y   = 0;
   }
   else {
      win_hgt = new_hgt;
      win_y   = (int)(disp_hgt - new_hgt) / 2;
      if (!win_y)
         win_y = 1;
   }
}

static void jump_request(CLK_VAL offset)
{
   if (sys.data.command == CMD_PAUSE)
      CONTROL_PLAY_ACTION(MFS_ENABLED);
   lock_MUTEX(sys.m, 1);
   while (sys.data.command & CMD_JUMP) {
      wait_CONDVAR(sys.cv, sys.m);
   }
   sys.jump_offset  = offset;
   sys.data.command = CMD_JUMP;
   unlock_MUTEX(sys.m);
}

static LRESULT CALLBACK event_handler(HWND hw, UINT msg, 
                                      WPARAM p1, LPARAM p2)
{
   switch (msg) {
   case WM_PAINT:
      PaintOverlay();
      break;

   case WM_COMMAND:
   {
      u32 last_cmd = sys.data.command;

      switch (LOWORD(p1)) {
      case ID_MAIN_MENU_CONTROL_PLAY:
         if (sys.data.command == CMD_SLOW_FWD)
            sys.data.command = CMD_PLAY;
         CONTROL_PLAY_ACTION(MFS_ENABLED);
         break;
      case ID_MAIN_MENU_CONTROL_SLOW:
         if (sys.data.command != CMD_SLOW_FWD) {
            if (sys.data.command == CMD_PAUSE)
               CONTROL_PLAY_ACTION(MFS_ENABLED);
            sys.data.command = CMD_SLOW_FWD;
         }
         else
            CONTROL_PLAY_ACTION(MFS_ENABLED);
         break;
      case ID_MAIN_MENU_CONTROL_JUMP_FWD:
         jump_request( CMD_JUMP_SECS*MPEG2_SYS_CLK_FREQ);
         break;
      case ID_MAIN_MENU_CONTROL_JUMP2_FWD:
         jump_request( CMD_JUMP2_SECS*MPEG2_SYS_CLK_FREQ);
         break;
      case ID_MAIN_MENU_CONTROL_JUMP_BCK:
         jump_request(-CMD_JUMP_SECS*MPEG2_SYS_CLK_FREQ);
         break;
      case ID_MAIN_MENU_CONTROL_JUMP2_BCK:
         jump_request(-CMD_JUMP2_SECS*MPEG2_SYS_CLK_FREQ);
         break;
      case ID_MAIN_MENU_CONTROL_JUMP2_START:
         jump_request( 0);
         break;
      case ID_MAIN_MENU_CONTROL_SELECTPROGRAM:
         win_mode         = WIN_MODE_SEL_PROG;
         sys.data.command = CMD_STOP;
         break;
      case ID_MAIN_MENU_CONTROL_SELECTSTREAMS:
         win_mode         = WIN_MODE_SEL_STR;
         sys.data.command = CMD_STOP;
         break;
      case ID_MAIN_MENU_FILE_OPEN:
         win_wid = win_hgt = 0;
         win_x   = win_y   = 0;
         FILE_OPEN_ACTION();
         break;
      case ID_MAIN_MENU_FILE_LOOP:
         FILE_LOOP_ACTION(1);
         break;
      case ID_MAIN_MENU_FILE_CLOSE:
         if ( (overlay)&&(win_mode == WIN_MODE_MAXIMIZED) )
            RestoreWindow();
         FILE_CLOSE_ACTION();
         break;
      case ID_MAIN_MENU_FILE_EXIT:
         goto EXIT_APP;
      case ID_USER_FULLSCREEN:
         if (overlay) {
            if (win_mode == WIN_MODE_MAXIMIZED)
               RestoreWindow();
            else
               PostMessage(win, WM_SIZE, SIZE_MAXIMIZED, 0);
         }
      }

      if ( (last_cmd == CMD_PAUSE)&&(sys.data.command != CMD_PAUSE) ) {
         broadcast_CONDVAR(sys.cv);
      }
      return 0;
   }
   case WM_SIZE:
      if (overlay) {
         switch (p1) {
         case SIZE_MAXIMIZED:
            if (win_mode != WIN_MODE_MAXIMIZED)
               MaximizeWindow();
            break;
         case SIZE_MINIMIZED:
            win_mode     = WIN_MODE_MINIMIZED;
            last_command = sys.data.command;
            if (sys.data.command != CMD_PAUSE) {
               /* pause before minimizing */
               sys.data.command = CMD_PLAY;
               CONTROL_PLAY_ACTION(MFS_ENABLED);
            }
            IDirectDrawSurface_UpdateOverlay(overlay, NULL, surface,
                                             NULL, DDOVER_HIDE, NULL);
            break;
         case SIZE_RESTORED:
            if (win_mode == WIN_MODE_MINIMIZED) {
               win_mode = WIN_MODE_NORMAL;
               if (last_command != CMD_PAUSE) {
                  /* unpause after restoring */
                  CONTROL_PLAY_ACTION(MFS_ENABLED);
                  sys.data.command = last_command;
                  broadcast_CONDVAR(sys.cv);
               }
            }
            get_win_size((u32)LOWORD(p2), (u32)HIWORD(p2));
            resize_DirectDrawOverlay();
         }
      }
      break;

   case WM_MOVE:
      if (overlay)
         resize_DirectDrawOverlay();
      break;

   case WM_SYSCOMMAND:
      if ( (overlay)&&(sys.data.command != CMD_PAUSE) ) {
         switch (p1 & 0xFFF0) {
         case SC_SCREENSAVE:
         case SC_MONITORPOWER:
            return 0;
         }
      }
      break;

   case WM_CLOSE:
EXIT_APP:
      if ( (overlay)&&(win_mode == WIN_MODE_MAXIMIZED) )
         RestoreWindow();
      FILE_CLOSE_ACTION();
      PostQuitMessage(0);
      return 0;
   }
   return DefWindowProc(hw, msg, p1, p2);
}

static int init_DirectDraw()
{
   const char   *err_msg;
/*   DDCAPS        caps;*/
   DDSURFACEDESC desc;

   if (DirectDrawCreate(NULL, &lpDD, NULL) != DD_OK) {
      err_msg = "Cannot initialize Direct Draw.";
      goto INIT_DD_ERR;
   }
/*
   memset(&caps, 0, sizeof(DDCAPS));
   caps.dwSize = sizeof(DDCAPS);
   if (IDirectDraw_GetCaps(lpDD, &caps, NULL) != DD_OK) {
      err_msg = "Cannot get Direct Draw capabilities.";
      goto INIT_DD_ERR;
   }
   else if ( (!(caps.dwCaps & DDCAPS_OVERLAY))||
             (!(caps.dwCaps & DDCAPS_OVERLAYSTRETCH))||
             (!(caps.dwCKeyCaps & DDCKEYCAPS_DESTOVERLAYYUV)) ) {
      err_msg = "Cannot use Direct Draw overlay.";
      goto INIT_DD_ERR;
   }
*/
   if (IDirectDraw_SetCooperativeLevel(lpDD, win, DDSCL_NORMAL) != DD_OK) {
      err_msg = "Cannot set Direct Draw co-operative level.";
      goto INIT_DD_ERR;
   }

   memset(&desc, 0, sizeof(DDSURFACEDESC));
   desc.dwSize         = sizeof(DDSURFACEDESC);
   desc.dwFlags        = DDSD_CAPS;
   desc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
   if (IDirectDraw_CreateSurface(lpDD, &desc, &surface, NULL) != DD_OK) {
      err_msg = "Cannot create a Direct Draw primary surface.";
      goto INIT_DD_ERR;
   }

#ifdef __USE_CLIPPER__
   /* Create a clipper object since this is for a Windowed render */
   if (IDirectDraw_CreateClipper(lpDD, 0, &clipper, NULL) != DD_OK) {
      err_msg = "Cannot create a clipper.";
      goto INIT_DD_ERR;
   }

   /* Associate the clipper with the window */
   IDirectDrawClipper_SetHWnd(clipper, 0, win);
   IDirectDrawSurface_SetClipper(surface, clipper);
#endif

   return 1;

INIT_DD_ERR:
   free_DirectDraw();
   MessageBox(win, err_msg,
              ERRMSG_TITLE, (MB_OK|MB_ICONEXCLAMATION));
   return 0;
}

static void free_DirectDraw()
{
#ifdef __USE_CLIPPER__
   if (clipper) {
      IDirectDrawClipper_Release(clipper);
      clipper = NULL;
   }
#endif

   if (surface) {
      IDirectDrawSurface_Release(surface);
      surface = NULL;
   }

   if (lpDD) {
      IDirectDraw_Release(lpDD);
      lpDD = NULL;
   }
}

static void get_overlay_colour()
{
   HDC           dc;
   COLORREF      rgbT;
   DDSURFACEDESC desc;
   HRESULT       rc;

   /*  Use GDI SetPixel to color match for us */
   rc = IDirectDrawSurface_GetDC(surface, &dc);
   if (!SUCCEEDED(rc))
      return;

   rgbT = GetPixel(dc, 0, 0);                /* Save current pixel value */
   SetPixel(dc, 0, 0, overlay_colour_24bit); /* Set our value */
   IDirectDrawSurface_ReleaseDC(surface, dc);

   /* Now lock the surface so we can read back the converted color */
   desc.dwSize = sizeof(DDSURFACEDESC);
   rc = IDirectDrawSurface_Lock(surface, NULL, &desc, DDLOCK_WAIT, NULL);
   if (!SUCCEEDED(rc))
      return;

   overlay_colour = *(u32 *)desc.lpSurface;
   if (desc.ddpfPixelFormat.dwRGBBitCount < 32)
      overlay_colour &= (1 << desc.ddpfPixelFormat.dwRGBBitCount) - 1;

   IDirectDrawSurface_Unlock(surface, NULL);

   /*  Now restore the original colour. */
   rc = IDirectDrawSurface_GetDC(surface, &dc);
   if (SUCCEEDED(rc)) {
      SetPixel(dc, 0, 0, rgbT);
      IDirectDrawSurface_ReleaseDC(surface, dc);
   }
}

static void PaintOverlay()
{
   PAINTSTRUCT paint;

   BeginPaint(win, &paint);
   if ( (overlay)&&(sys.data.command != CMD_STOP) ) {
      HBRUSH brush;
      RECT   ovr_rect, paint_rect,
            *paint_area = &paint.rcPaint;

      if ( (win_x)||(win_y) ) {
         brush = CreateSolidBrush(GetNearestColor(paint.hdc, 
                                                  RGB(0, 0, 0)));
         if (GetClientRect(win, &ovr_rect)) {
            ovr_rect.left  += (long)win_x;
            ovr_rect.top   += (long)win_y;
            ovr_rect.right  = ovr_rect.left + (long)win_wid;
            ovr_rect.bottom = ovr_rect.top  + (long)win_hgt;

            if (IntersectRect(&paint_rect, paint_area, &ovr_rect)) {
               FillRect(paint.hdc, paint_area, brush);
               DeleteObject(brush);

               brush = CreateSolidBrush(GetNearestColor(paint.hdc, 
                                                        overlay_colour_24bit));
               paint_area = &paint_rect;
            }
         }
      }
      else {
         brush = CreateSolidBrush(GetNearestColor(paint.hdc, 
                                                  overlay_colour_24bit));
      }

      FillRect(paint.hdc, paint_area, brush);
      DeleteObject(brush);
   }
   else
      FillRect(paint.hdc, &paint.rcPaint, (HBRUSH)COLOR_WINDOW);
   EndPaint(win, &paint);
}

static void RestoreWindow()
{
   win_mode = WIN_MODE_NORMAL;
   SetWindowLong(win, GWL_STYLE, (DWSTYLE | WS_VISIBLE)); 
   SetMenu(win, win_menu);
   SetWindowPos(win, HWND_TOP, 
                (int)last_pos.left, (int)last_pos.top,
                (int)(last_pos.right - last_pos.left),
                (int)(last_pos.bottom - last_pos.top),
                SWP_SHOWWINDOW);
   ShowCursor(TRUE);
}

static void MaximizeWindow()
{
   win_mode = WIN_MODE_MAXIMIZED;
         
   GetWindowRect(win, &last_pos);
   SetWindowLong(win, GWL_STYLE, WS_VISIBLE); 
   SetWindowPos(win, HWND_TOPMOST, 0, 0,
                (int)screen_wid, (int)screen_hgt,
                SWP_SHOWWINDOW);
   SetMenu(win, NULL);
   ShowCursor(FALSE);
}


#undef __USE_CLIPPER__

